import unittest

from pypika import (
    SYSTEM_TIME,
    Field,
    Interval,
    JoinException,
    JoinType,
    MySQLQuery,
    Query,
    SetOperationException,
    Table,
    Tables,
)
from pypika import (
    functions as fn,
)

__author__ = "Timothy Heys"
__email__ = "theys@kayak.com"


class SelectQueryJoinTests(unittest.TestCase):
    table0, table1, hij = Tables("abc", "efg", "hij")

    def test_default_join_type(self):
        query = Query.from_(self.table0).join(self.table1).on(self.table0.foo == self.table1.bar).select("*")

        self.assertEqual('SELECT * FROM "abc" JOIN "efg" ON "abc"."foo"="efg"."bar"', str(query))

    def test_inner_join(self):
        expected = 'SELECT * FROM "abc" JOIN "efg" ON "abc"."foo"="efg"."bar"'

        with self.subTest("join with enum"):
            query = (
                Query.from_(self.table0)
                .join(self.table1, how=JoinType.inner)
                .on(self.table0.foo == self.table1.bar)
                .select("*")
            )
            self.assertEqual(expected, str(query))

        with self.subTest("join function"):
            query = Query.from_(self.table0).inner_join(self.table1).on(self.table0.foo == self.table1.bar).select("*")
            self.assertEqual(expected, str(query))

    def test_left_join(self):
        expected = 'SELECT * FROM "abc" LEFT JOIN "efg" ON "abc"."foo"="efg"."bar"'

        with self.subTest("join with enum"):
            query = (
                Query.from_(self.table0)
                .join(self.table1, how=JoinType.left)
                .on(self.table0.foo == self.table1.bar)
                .select("*")
            )

            self.assertEqual(expected, str(query))

        with self.subTest("join function"):
            query = Query.from_(self.table0).left_join(self.table1).on(self.table0.foo == self.table1.bar).select("*")
            self.assertEqual(expected, str(query))

    def test_right_join(self):
        expected = 'SELECT * FROM "abc" RIGHT JOIN "efg" ON "abc"."foo"="efg"."bar"'

        with self.subTest("join with enum"):
            query = (
                Query.from_(self.table0)
                .join(self.table1, how=JoinType.right)
                .on(self.table0.foo == self.table1.bar)
                .select("*")
            )

            self.assertEqual(expected, str(query))

        with self.subTest("join function"):
            query = Query.from_(self.table0).right_join(self.table1).on(self.table0.foo == self.table1.bar).select("*")
            self.assertEqual(expected, str(query))

    def test_hash_join(self):
        expected = 'SELECT * FROM "abc" HASH JOIN "efg" ON "abc"."foo"="efg"."bar"'

        with self.subTest("join with enum"):
            query = (
                Query.from_(self.table0)
                .join(self.table1, how=JoinType.hash)
                .on(self.table0.foo == self.table1.bar)
                .select("*")
            )

            self.assertEqual(expected, str(query))

        with self.subTest("join function"):
            query = Query.from_(self.table0).hash_join(self.table1).on(self.table0.foo == self.table1.bar).select("*")
            self.assertEqual(expected, str(query))

    def test_outer_join(self):
        expected = 'SELECT * FROM "abc" FULL OUTER JOIN "efg" ON "abc"."foo"="efg"."bar"'

        with self.subTest("join with enum"):
            query = (
                Query.from_(self.table0)
                .join(self.table1, how=JoinType.outer)
                .on(self.table0.foo == self.table1.bar)
                .select("*")
            )

            self.assertEqual(expected, str(query))

        with self.subTest("join function"):
            query = Query.from_(self.table0).outer_join(self.table1).on(self.table0.foo == self.table1.bar).select("*")
            self.assertEqual(expected, str(query))

    def test_cross_join(self):
        expected = 'SELECT * FROM "abc" CROSS JOIN "efg" ON "abc"."foo"="efg"."bar"'

        with self.subTest("join with enum"):
            query = (
                Query.from_(self.table0)
                .join(self.table1, how=JoinType.cross)
                .on(self.table0.foo == self.table1.bar)
                .select("*")
            )

            self.assertEqual(expected, str(query))

        with self.subTest("join function"):
            query = Query.from_(self.table0).cross_join(self.table1).on(self.table0.foo == self.table1.bar).select("*")
            self.assertEqual(expected, str(query))

    def test_left_outer_join(self):
        expected = 'SELECT * FROM "abc" LEFT OUTER JOIN "efg" ON "abc"."foo"="efg"."bar"'
        with self.subTest("join with enum"):
            query = (
                Query.from_(self.table0)
                .join(self.table1, how=JoinType.left_outer)
                .on(self.table0.foo == self.table1.bar)
                .select("*")
            )

            self.assertEqual(expected, str(query))

        with self.subTest("join function"):
            query = (
                Query.from_(self.table0).left_outer_join(self.table1).on(self.table0.foo == self.table1.bar).select("*")
            )

            self.assertEqual(expected, str(query))

    def test_right_outer_join(self):
        expected = 'SELECT * FROM "abc" RIGHT OUTER JOIN "efg" ON "abc"."foo"="efg"."bar"'
        with self.subTest("join with enum"):
            query = (
                Query.from_(self.table0)
                .join(self.table1, how=JoinType.right_outer)
                .on(self.table0.foo == self.table1.bar)
                .select("*")
            )

            self.assertEqual(expected, str(query))

        with self.subTest("join function"):
            query = (
                Query.from_(self.table0)
                .right_outer_join(self.table1)
                .on(self.table0.foo == self.table1.bar)
                .select("*")
            )

            self.assertEqual(expected, str(query))

    def test_full_outer_join(self):
        expected = 'SELECT * FROM "abc" FULL OUTER JOIN "efg" ON "abc"."foo"="efg"."bar"'

        with self.subTest("join with enum"):
            query = (
                Query.from_(self.table0)
                .join(self.table1, how=JoinType.full_outer)
                .on(self.table0.foo == self.table1.bar)
                .select("*")
            )

            self.assertEqual(expected, str(query))

        with self.subTest("join function"):
            query = (
                Query.from_(self.table0).full_outer_join(self.table1).on(self.table0.foo == self.table1.bar).select("*")
            )

            self.assertEqual(expected, str(query))

    def test_join_on_field_single(self):
        query = Query.from_(self.table0).join(self.table1).on_field("foo").select("*")
        self.assertEqual('SELECT * FROM "abc" JOIN "efg" ON "abc"."foo"="efg"."foo"', str(query))

    def test_join_on_field_multi(self):
        query = Query.from_(self.table0).join(self.table1).on_field("foo", "bar").select("*")
        self.assertEqual(
            'SELECT * FROM "abc" JOIN "efg" ON "abc"."foo"="efg"."foo" AND "abc"."bar"="efg"."bar"',
            str(query),
        )

    def test_join_on_field_multi_with_extra_join(self):
        query = (
            Query.from_(self.table0)
            .join(self.hij)
            .on_field("buzz")
            .join(self.table1)
            .on_field("foo", "bar")
            .select("*")
        )

        self.assertEqual(
            'SELECT * FROM "abc" JOIN "hij" ON "abc"."buzz"="hij"."buzz" '
            'JOIN "efg" ON "abc"."foo"="efg"."foo" AND "abc"."bar"="efg"."bar"',
            str(query),
        )

    def test_join_using_string_field_name(self):
        query = Query.from_(self.table0).join(self.table1).using("id").select("*")

        self.assertEqual('SELECT * FROM "abc" JOIN "efg" USING ("id")', str(query))

    def test_join_using_multiple_fields(self):
        query = Query.from_(self.table0).join(self.table1).using("foo", "bar").select("*")

        self.assertEqual('SELECT * FROM "abc" JOIN "efg" USING ("foo","bar")', str(query))

    def test_join_using_with_quote_char(self):
        query = Query.from_(self.table0).join(self.table1).using("foo", "bar").select("*")

        self.assertEqual("SELECT * FROM abc JOIN efg USING (foo,bar)", query.get_sql(quote_char=""))

    def test_join_using_without_fields_raises_exception(self):
        with self.assertRaises(JoinException):
            Query.from_(self.table0).join(self.table1).using()

    def test_join_on_field_without_fields_raises_exception(self):
        with self.assertRaises(JoinException):
            Query.from_(self.table0).join(self.table1).on_field()

    def test_join_arithmetic_field(self):
        q = (
            Query.from_(self.table0)
            .join(self.table1)
            .on(self.table0.dt == (self.table1.dt - Interval(weeks=1)))
            .select("*")
        )

        self.assertEqual(
            'SELECT * FROM "abc" JOIN "efg" ON "abc"."dt"="efg"."dt"-INTERVAL \'1 WEEK\'',
            str(q),
        )

    def test_join_with_arithmetic_function_in_select(self):
        q = (
            Query.from_(
                self.table0,
            )
            .join(self.table1)
            .on(self.table0.dt == (self.table1.dt - Interval(weeks=1)))
            .select(self.table0.fiz - self.table0.buz, self.table1.star)
        )

        self.assertEqual(
            'SELECT "abc"."fiz"-"abc"."buz","efg".* FROM "abc" JOIN "efg" ON "abc"."dt"="efg"."dt"-INTERVAL \'1 WEEK\'',
            str(q),
        )

    def test_join_on_complex_criteria(self):
        q = (
            Query.from_(self.table0)
            .join(self.table1, how=JoinType.right)
            .on((self.table0.foo == self.table1.fiz) & (self.table0.bar == self.table1.buz))
            .select("*")
        )

        self.assertEqual(
            'SELECT * FROM "abc" RIGHT JOIN "efg" ON "abc"."foo"="efg"."fiz" AND "abc"."bar"="efg"."buz"',
            str(q),
        )

    def test_join_on_subquery_criteria(self):
        table_a, table_b, table_c = Tables("a", "b", "c")
        subquery = Query.from_(table_c).select("id").limit(1)
        query = (
            Query.from_(table_a).select("*").join(table_b).on((table_a.b_id == table_b.id) & (table_b.c_id == subquery))
        )

        self.assertEqual(
            "SELECT * "
            'FROM "a" '
            'JOIN "b" ON "a"."b_id"="b"."id" AND "b"."c_id"='
            '(SELECT "id" FROM "c" LIMIT 1)',
            str(query),
        )

    def test_use_different_table_objects_for_same_table(self):
        table = Table("t")
        q = Query.from_(table).select("*").where(Field("id", table=table) == 1)

        self.assertEqual('SELECT * FROM "t" WHERE "id"=1', str(q))

    def test_join_second_table_in_from_clause(self):
        table_a, table_b, table_c = Tables("a", "b", "c")
        q = Query.from_(table_a).from_(table_b).select("*").join(table_c).on(table_b.c_id == table_c.id)

        self.assertEqual("SELECT * FROM \"a\",\"b\" JOIN \"c\" ON \"b\".\"c_id\"=\"c\".\"id\"", str(q))

    def test_cross_join_on_table(self):
        table_a, table_b = Tables("a", "b")
        q = Query.from_(table_a).join(table_b).cross().select("*")

        self.assertEqual('SELECT * FROM "a" CROSS JOIN "b"', str(q))

    def test_cross_join_on_subquery(self):
        table_a, table_b = Tables("a", "b")
        q_a = Query.from_(table_a).select("*")
        q_b = Query.from_(table_b).select("*").join(q_a).cross().select("*")

        self.assertEqual('SELECT * FROM "b" CROSS JOIN (SELECT * FROM "a") "sq0"', str(q_b))

    def test_join_on_collate(self):
        table_a, table_b = Tables("a", "b")

        q1 = (
            Query.from_(table_a)
            .select(table_b.ouch)
            .join(table_b)
            .on(table_a.foo == table_b.boo, collate="utf8_general_ci")
        )
        q2 = Query.from_(table_a).select(table_b.ouch).join(table_b).on(table_a.foo == table_b.boo)

        self.assertEqual(
            'SELECT "b"."ouch" FROM "a" JOIN "b" ON "a"."foo"="b"."boo" COLLATE utf8_general_ci',
            str(q1),
        )
        self.assertEqual('SELECT "b"."ouch" FROM "a" JOIN "b" ON "a"."foo"="b"."boo"', str(q2))

    def test_temporal_join(self):
        t0 = self.table0.for_(SYSTEM_TIME.as_of('2020-01-01'))
        t1 = self.table1.for_(SYSTEM_TIME.as_of('2020-01-01'))
        query = Query.from_(t0).join(t1).on(t0.foo == t1.bar).select("*")

        self.assertEqual(
            'SELECT * FROM "abc" FOR SYSTEM_TIME AS OF \'2020-01-01\' '
            'JOIN "efg" FOR SYSTEM_TIME AS OF \'2020-01-01\' '
            'ON "abc"."foo"="efg"."bar"',
            str(query),
        )


class JoinBehaviorTests(unittest.TestCase):
    table_abc, table_efg, table_hij, table_klm = Tables("abc", "efg", "hij", "klm")

    def test_select__ordered_select_clauses(self):
        q = (
            Query.from_(self.table_abc)
            .join(self.table_efg)
            .on(self.table_abc.foo == self.table_efg.bar)
            .select(
                self.table_abc.baz,
                self.table_efg.buz,
                self.table_abc.fiz,
                self.table_efg.bam,
            )
        )

        self.assertEqual(
            'SELECT "abc"."baz","efg"."buz","abc"."fiz","efg"."bam" FROM "abc" JOIN "efg" ON "abc"."foo"="efg"."bar"',
            str(q),
        )

    def test_select__star_for_table(self):
        q = (
            Query.from_(self.table_abc)
            .join(self.table_efg)
            .on(self.table_abc.foo == self.table_efg.bar)
            .join(self.table_hij)
            .on(self.table_abc.buz == self.table_hij.bam)
            .select(self.table_abc.star)
            .select(self.table_efg.star)
            .select(self.table_hij.star)
        )

        self.assertEqual(
            'SELECT "abc".*,"efg".*,"hij".* FROM "abc" '
            'JOIN "efg" ON "abc"."foo"="efg"."bar" '
            'JOIN "hij" ON "abc"."buz"="hij"."bam"',
            str(q),
        )

    def test_select__star_for_table__replacement(self):
        q = (
            Query.from_(self.table_abc)
            .join(self.table_efg)
            .on(self.table_abc.foo == self.table_efg.bar)
            .join(self.table_hij)
            .on(self.table_abc.buz == self.table_hij.bam)
            .select(self.table_abc.foo, self.table_efg.bar, self.table_hij.bam)
            .select(self.table_abc.star, self.table_efg.star, self.table_hij.star)
        )

        self.assertEqual(
            'SELECT "abc".*,"efg".*,"hij".* FROM "abc" '
            'JOIN "efg" ON "abc"."foo"="efg"."bar" '
            'JOIN "hij" ON "abc"."buz"="hij"."bam"',
            str(q),
        )

    def test_select_fields_with_where(self):
        q = (
            Query.from_(self.table_abc)
            .join(self.table_efg)
            .on(self.table_abc.foo == self.table_efg.bar)
            .join(self.table_hij)
            .on(self.table_abc.buz == self.table_hij.bam)
            .select(self.table_abc.foo, self.table_efg.bar, self.table_hij.bam)
            .where(self.table_abc.foo > 1)
            .where(self.table_efg.bar != 2)
        )

        self.assertEqual(
            'SELECT "abc"."foo","efg"."bar","hij"."bam" FROM "abc" '
            'JOIN "efg" ON "abc"."foo"="efg"."bar" '
            'JOIN "hij" ON "abc"."buz"="hij"."bam" '
            'WHERE "abc"."foo">1 AND "efg"."bar"<>2',
            str(q),
        )

    def test_require_condition(self):
        with self.assertRaises(JoinException):
            Query.from_(self.table_abc).join(self.table_efg).on(None)

    def test_require_condition_with_both_tables(self):
        with self.assertRaises(JoinException):
            Query.from_(self.table_abc).join(self.table_efg).on(self.table_abc.foo == self.table_hij.bar)

        with self.assertRaises(JoinException):
            Query.from_(self.table_abc).join(self.table_efg).on(self.table_hij.foo == self.table_efg.bar)

        with self.assertRaises(JoinException):
            Query.from_(self.table_abc).join(self.table_efg).on(self.table_hij.foo == self.table_klm.bar)

    def test_join_same_table(self):
        table1 = Table("abc")
        table2 = Table("abc")
        q = Query.from_(table1).join(table2).on(table1.foo == table2.bar).select(table1.foo, table2.buz)

        self.assertEqual(
            'SELECT "abc"."foo","abc2"."buz" FROM "abc" JOIN "abc" "abc2" ON "abc"."foo"="abc2"."bar"',
            str(q),
        )

    def test_join_same_table_with_prefixes(self):
        table1 = Table("abc").as_("x")
        table2 = Table("abc").as_("y")
        q = Query.from_(table1).join(table2).on(table1.foo == table2.bar).select(table1.foo, table2.buz)

        self.assertEqual(
            'SELECT "x"."foo","y"."buz" FROM "abc" "x" JOIN "abc" "y" ON "x"."foo"="y"."bar"',
            str(q),
        )

    def test_join_table_twice(self):
        table1, table2 = Table("efg").as_("efg1"), Table("efg").as_("efg2")
        q = (
            Query.from_(self.table_abc)
            .join(table1)
            .on(self.table_abc.foo == table1.bar)
            .join(table2)
            .on(self.table_abc.foo == table2.bam)
            .select(self.table_abc.foo, table1.fiz, table2.buz)
        )

        self.assertEqual(
            'SELECT "abc"."foo","efg1"."fiz","efg2"."buz" FROM "abc" '
            'JOIN "efg" "efg1" ON "abc"."foo"="efg1"."bar" '
            'JOIN "efg" "efg2" ON "abc"."foo"="efg2"."bam"',
            str(q),
        )

    def test_select__fields_after_table_star(self):
        q = (
            Query.from_(self.table_abc)
            .join(self.table_efg)
            .on(self.table_abc.foo == self.table_efg.bar)
            .select(self.table_abc.star, self.table_efg.bar)
            .select(self.table_abc.foo)
        )

        self.assertEqual(
            'SELECT "abc".*,"efg"."bar" FROM "abc" JOIN "efg" ON "abc"."foo"="efg"."bar"',
            str(q),
        )

    def test_fail_when_joining_unknown_type(self):
        with self.assertRaises(ValueError):
            Query.from_(self.table_abc).join("this is a string")

    def test_immutable__tables(self):
        query1 = Query.from_(self.table_abc).select(self.table_abc.foo)
        query2 = (
            Query.from_(self.table_abc)
            .join(self.table_efg)
            .on(self.table_abc.foo == self.table_efg.bar)
            .select(self.table_abc.foo, self.table_efg.buz)
        )

        self.assertEqual(
            'SELECT "abc"."foo","efg"."buz" FROM "abc" JOIN "efg" ON "abc"."foo"="efg"."bar"',
            str(query2),
        )
        self.assertEqual('SELECT "foo" FROM "abc"', str(query1))

    def test_select_field_from_missing_table(self):
        """
        Previously we validated the select field was a member of a table
        in the query (e.g. FROM clause or a JOIN). However this made
        correlated subqueries awkward to write (subqueries which refer
        to their parent table. So the validation was removed.

        None of the following raise an error. Previously raised JoinException
        """
        Query.from_(self.table_abc).select(self.table_efg.foo)

        Query.from_(self.table_abc).where(self.table_efg.foo == 0)

        Query.from_(self.table_abc).where(fn.Sum(self.table_efg.foo) == 0)

        Query.from_(self.table_abc).select(fn.Sum(self.table_abc.bar * 2) + fn.Sum(self.table_efg.foo * 2))

        Query.from_(self.table_abc).groupby(self.table_efg.foo)

        Query.from_(self.table_abc).groupby(self.table_abc.foo).having(self.table_efg.bar)

        subquery = (
            Query.from_(self.table_efg).select(self.table_efg.id).where(self.table_efg.abc_id == self.table_abc.id)
        )
        query = Query.from_(self.table_abc).select(subquery.as_("efg_id").limit(1))
        self.assertEqual(
            'SELECT (SELECT "efg"."id" FROM "efg" WHERE "efg"."abc_id"="abc"."id" LIMIT 1) "efg_id" FROM "abc"',
            str(query),
        )

    def test_ignore_table_references(self):
        query = Query.from_(Table("abc")).select(Table("abc").foo)

        self.assertEqual('SELECT "foo" FROM "abc"', str(query))

    def test_prefixes_added_to_groupby(self):
        test_query = (
            Query.from_(self.table_abc)
            .join(self.table_efg)
            .on(self.table_abc.foo == self.table_efg.bar)
            .select(self.table_abc.foo, fn.Sum(self.table_efg.buz))
            .groupby(self.table_abc.foo)
        )

        self.assertEqual(
            'SELECT "abc"."foo",SUM("efg"."buz") FROM "abc" JOIN "efg" ON "abc"."foo"="efg"."bar" GROUP BY "abc"."foo"',
            str(test_query),
        )

    def test_prefixes_added_to_orderby(self):
        test_query = (
            Query.from_(self.table_abc)
            .join(self.table_efg)
            .on(self.table_abc.foo == self.table_efg.bar)
            .select(self.table_abc.foo, self.table_efg.buz)
            .orderby(self.table_abc.foo)
        )

        self.assertEqual(
            'SELECT "abc"."foo","efg"."buz" FROM "abc" JOIN "efg" ON "abc"."foo"="efg"."bar" ORDER BY "abc"."foo"',
            str(test_query),
        )

    def test_prefixes_added_to_function_in_orderby(self):
        test_query = (
            Query.from_(self.table_abc)
            .join(self.table_efg)
            .on(self.table_abc.foo == self.table_efg.bar)
            .select(self.table_abc.foo, self.table_efg.buz)
            .orderby(fn.Date(self.table_abc.foo))
        )

        self.assertEqual(
            'SELECT "abc"."foo","efg"."buz" FROM "abc" '
            'JOIN "efg" ON "abc"."foo"="efg"."bar" '
            'ORDER BY DATE("abc"."foo")',
            str(test_query),
        )

    def test_join_from_join(self):
        test_query = (
            Query.from_(self.table_abc)
            .join(self.table_efg)
            .on(self.table_abc.efg_id == self.table_efg.id)
            .join(self.table_hij)
            .on(self.table_efg.hij_id == self.table_hij.id)
            .select(self.table_abc.foo, self.table_efg.bar, self.table_hij.fizz)
        )

        self.assertEqual(
            'SELECT "abc"."foo","efg"."bar","hij"."fizz" FROM "abc" '
            'JOIN "efg" ON "abc"."efg_id"="efg"."id" '
            'JOIN "hij" ON "efg"."hij_id"="hij"."id"',
            str(test_query),
        )

    def test_join_query_without_alias(self):
        subquery = Query.from_(self.table_efg).select(
            self.table_efg.base_id.as_("x"),
            self.table_efg.fizz,
            self.table_efg.buzz,
        )

        test_query = (
            Query.from_(self.table_abc)
            .join(subquery)
            .on(subquery.x == self.table_abc.id)
            .select(self.table_abc.foo, subquery.fizz, subquery.buzz)
        )

        self.assertEqual(
            'SELECT "abc"."foo","sq0"."fizz","sq0"."buzz" '
            'FROM "abc" JOIN ('
            'SELECT "base_id" "x","fizz","buzz" FROM "efg"'
            ') "sq0" '
            'ON "sq0"."x"="abc"."id"',
            str(test_query),
        )

    def test_join_query_with_column_alias(self):
        subquery = (
            Query.from_(self.table_efg)
            .select(
                self.table_efg.base_id.as_("x"),
                self.table_efg.fizz,
                self.table_efg.buzz,
            )
            .as_("subq")
        )

        test_query = (
            Query.from_(self.table_abc)
            .join(subquery)
            .on(subquery.x == self.table_abc.id)
            .select(self.table_abc.foo, subquery.fizz, subquery.buzz)
        )

        self.assertEqual(
            'SELECT "abc"."foo","subq"."fizz","subq"."buzz" '
            'FROM "abc" JOIN ('
            'SELECT "base_id" "x","fizz","buzz" FROM "efg"'
            ') "subq" '
            'ON "subq"."x"="abc"."id"',
            str(test_query),
        )

    def test_join_query_with_table_alias(self):
        xxx = self.table_efg.as_("xxx")
        subquery = Query.from_(xxx).select(xxx.base_id, xxx.fizz, xxx.buzz).as_("subq")

        test_query = (
            Query.from_(self.table_abc)
            .join(subquery)
            .on(subquery.x == self.table_abc.id)
            .select(self.table_abc.foo, subquery.fizz, subquery.buzz)
        )

        self.assertEqual(
            'SELECT "abc"."foo","subq"."fizz","subq"."buzz" '
            'FROM "abc" JOIN ('
            'SELECT "xxx"."base_id","xxx"."fizz","xxx"."buzz" FROM "efg" "xxx"'
            ') "subq" '
            'ON "subq"."x"="abc"."id"',
            str(test_query),
        )

    def test_join_query_with_setoperation(self):
        subquery = Query.from_(self.table_abc).select("*").union(Query.from_(self.table_abc).select("*")).as_("subq")

        test_query = (
            Query.from_(self.table_abc).join(subquery).on(subquery.x == self.table_abc.id).select(self.table_abc.foo)
        )

        self.assertEqual(
            'SELECT "abc"."foo" FROM "abc" '
            'JOIN '
            '((SELECT * FROM "abc") '
            'UNION '
            '(SELECT * FROM "abc")) "subq" '
            'ON "subq"."x"="abc"."id"',
            str(test_query),
        )


class UnionTests(unittest.TestCase):
    table1, table2 = Tables("abc", "efg")

    def test_union(self):
        query1 = Query.from_(self.table1).select(self.table1.foo)
        query2 = Query.from_(self.table2).select(self.table2.bar)

        self.assertEqual(
            '(SELECT "foo" FROM "abc") UNION (SELECT "bar" FROM "efg")',
            str(query1 + query2),
        )
        self.assertEqual(
            '(SELECT "foo" FROM "abc") UNION (SELECT "bar" FROM "efg")',
            str(query1.union(query2)),
        )

    def test_union_multiple(self):
        table3, table4 = Tables("hij", "lmn")
        query1 = Query.from_(self.table1).select(self.table1.foo)
        query2 = Query.from_(self.table2).select(self.table2.bar)
        query3 = Query.from_(table3).select(table3.baz)
        query4 = Query.from_(table4).select(table4.faz)

        self.assertEqual(
            '(SELECT "foo" FROM "abc") UNION '
            '(SELECT "bar" FROM "efg") UNION '
            '(SELECT "baz" FROM "hij") UNION '
            '(SELECT "faz" FROM "lmn")',
            str(query1 + query2 + query3 + query4),
        )
        self.assertEqual(
            '(SELECT "foo" FROM "abc") UNION '
            '(SELECT "bar" FROM "efg") UNION '
            '(SELECT "baz" FROM "hij") UNION '
            '(SELECT "faz" FROM "lmn")',
            str(query1.union(query2).union(query3).union(query4)),
        )

    def test_union_all(self):
        query1 = Query.from_(self.table1).select(self.table1.foo)
        query2 = Query.from_(self.table2).select(self.table2.bar)

        self.assertEqual(
            '(SELECT "foo" FROM "abc") UNION ALL (SELECT "bar" FROM "efg")',
            str(query1 * query2),
        )
        self.assertEqual(
            '(SELECT "foo" FROM "abc") UNION ALL (SELECT "bar" FROM "efg")',
            str(query1.union_all(query2)),
        )

    def test_union_all_multiple(self):
        table3, table4 = Tables("hij", "lmn")
        query1 = Query.from_(self.table1).select(self.table1.foo)
        query2 = Query.from_(self.table2).select(self.table2.bar)
        query3 = Query.from_(table3).select(table3.baz)
        query4 = Query.from_(table4).select(table4.faz)

        self.assertEqual(
            '(SELECT "foo" FROM "abc") UNION ALL '
            '(SELECT "bar" FROM "efg") UNION ALL '
            '(SELECT "baz" FROM "hij") UNION ALL '
            '(SELECT "faz" FROM "lmn")',
            str(query1 * query2 * query3 * query4),
        )
        self.assertEqual(
            '(SELECT "foo" FROM "abc") UNION ALL '
            '(SELECT "bar" FROM "efg") UNION ALL '
            '(SELECT "baz" FROM "hij") UNION ALL '
            '(SELECT "faz" FROM "lmn")',
            str(query1.union_all(query2).union_all(query3).union_all(query4)),
        )

    def test_union_with_order_by(self):
        query1 = Query.from_(self.table1).select(self.table1.foo.as_("a"))
        query2 = Query.from_(self.table2).select(self.table2.bar.as_("a"))

        union_query = str((query1 + query2).orderby(query1.field("a")))
        union_all_query = str((query1 * query2).orderby(query1.field("a")))

        self.assertEqual(
            '(SELECT "foo" "a" FROM "abc") UNION (SELECT "bar" "a" FROM "efg") ORDER BY "a"',
            union_query,
        )

        self.assertEqual(
            '(SELECT "foo" "a" FROM "abc") UNION ALL (SELECT "bar" "a" FROM "efg") ORDER BY "a"',
            union_all_query,
        )

    def test_union_with_order_by_use_union_query_field(self):
        query1 = Query.from_(self.table1).select(self.table1.foo.as_("a"))
        query2 = Query.from_(self.table2).select(self.table2.bar.as_("a"))

        union_query = query1 + query2
        union_query = str(union_query.orderby(union_query.field("a")))
        union_all_query = query1 * query2
        union_all_query = str(union_all_query.orderby(union_all_query.field("a")))

        self.assertEqual(
            '(SELECT "foo" "a" FROM "abc") UNION (SELECT "bar" "a" FROM "efg") ORDER BY "a"',
            union_query,
        )
        self.assertEqual(
            '(SELECT "foo" "a" FROM "abc") UNION ALL (SELECT "bar" "a" FROM "efg") ORDER BY "a"',
            union_all_query,
        )

    def test_union_with_order_by_with_aliases(self):
        query1 = Query.from_(self.table1.as_("a")).select(self.table1.foo.as_("a"))
        query2 = Query.from_(self.table2.as_("b")).select(self.table2.bar.as_("a"))

        union_all_query = str((query1 * query2).orderby(query1.field("a")))
        union_query = str((query1 + query2).orderby(query1.field("a")))
        self.assertEqual(
            '(SELECT "foo" "a" FROM "abc" "a") UNION ALL (SELECT "bar" "a" FROM "efg" "b") ORDER BY "a"',
            union_all_query,
        )
        self.assertEqual(
            '(SELECT "foo" "a" FROM "abc" "a") UNION (SELECT "bar" "a" FROM "efg" "b") ORDER BY "a"',
            union_query,
        )

    def test_union_with_order_by_use_union_query_field_with_aliases(self):
        query1 = Query.from_(self.table1).select(self.table1.foo.as_("a"))
        query2 = Query.from_(self.table2).select(self.table2.bar.as_("a"))

        union_query = (query1 + query2).as_("x")
        union_query = union_query.orderby(union_query.field("a"))
        union_all_query = (query1 * query2).as_("x")
        union_all_query = union_all_query.orderby(union_all_query.field("a"))

        self.assertEqual(
            '(SELECT "foo" "a" FROM "abc") UNION (SELECT "bar" "a" FROM "efg") ORDER BY "x"."a"',
            str(union_query),
        )
        self.assertEqual(
            '(SELECT "foo" "a" FROM "abc") UNION ALL (SELECT "bar" "a" FROM "efg") ORDER BY "x"."a"',
            str(union_all_query),
        )

    def test_require_equal_number_of_fields(self):
        query1 = Query.from_(self.table1).select(self.table1.foo)
        query2 = Query.from_(self.table2).select(self.table2.fiz, self.table2.buz)

        with self.assertRaises(SetOperationException):
            str(query1 + query2)

    def test_mysql_query_wraps_unioned_queries(self):
        query1 = MySQLQuery.from_(self.table1).select(self.table1.foo)
        query2 = Query.from_(self.table2).select(self.table2.bar)

        self.assertEqual(
            "(SELECT `foo` FROM `abc`) UNION (SELECT `bar` FROM `efg`)",
            str(query1 + query2),
        )

    def test_union_as_subquery(self):
        abc, efg = Tables("abc", "efg")
        hij = Query.from_(abc).select(abc.t).union(Query.from_(efg).select(efg.t))
        q = Query.from_(hij).select(fn.Avg(hij.t))

        self.assertEqual(
            'SELECT AVG("sq0"."t") FROM ((SELECT "t" FROM "abc") UNION (SELECT "t" FROM "efg")) "sq0"',
            str(q),
        )

    def test_union_with_no_quote_char(self):
        abc, efg = Tables("abc", "efg")
        hij = Query.from_(abc).select(abc.t).union(Query.from_(efg).select(efg.t))
        q = Query.from_(hij).select(fn.Avg(hij.t))

        self.assertEqual(
            "SELECT AVG(sq0.t) FROM ((SELECT t FROM abc) UNION (SELECT t FROM efg)) sq0",
            q.get_sql(quote_char=None),
        )


class InsertQueryJoinTests(unittest.TestCase):
    def test_join_table_on_insert_query(self):
        a, b, c = Tables("a", "b", "c")
        q = Query.into(c).from_(a).join(b).on(a.fkey_id == b.id).where(b.foo == 1).select("a.*", "b.*")

        self.assertEqual(
            'INSERT INTO "c" SELECT "a"."a.*","a"."b.*" FROM "a" JOIN "b" ON "a"."fkey_id"="b"."id" WHERE "b"."foo"=1',
            str(q),
        )


class UpdateQueryJoinTests(unittest.TestCase):
    def test_join_table_on_update_query(self):
        a, b = Tables("a", "b")
        q = Query.update(a).join(b).on(a.fkey_id == b.id).where(b.foo == 1).set("adwords_batch_job_id", 1)

        self.assertEqual(
            'UPDATE "a" JOIN "b" ON "a"."fkey_id"="b"."id" SET "adwords_batch_job_id"=1 WHERE "b"."foo"=1',
            str(q),
        )


class IntersectTests(unittest.TestCase):
    table1, table2 = Tables("abc", "efg")

    def test_intersect(self):
        query1 = Query.from_(self.table1).select(self.table1.foo)
        query2 = Query.from_(self.table2).select(self.table2.bar)

        self.assertEqual(
            '(SELECT "foo" FROM "abc") INTERSECT (SELECT "bar" FROM "efg")',
            str(query1.intersect(query2)),
        )

    def test_intersect_multiple(self):
        table3, table4 = Tables("hij", "lmn")
        query1 = Query.from_(self.table1).select(self.table1.foo)
        query2 = Query.from_(self.table2).select(self.table2.bar)
        query3 = Query.from_(table3).select(table3.baz)
        query4 = Query.from_(table4).select(table4.faz)

        self.assertEqual(
            '(SELECT "foo" FROM "abc") INTERSECT '
            '(SELECT "bar" FROM "efg") INTERSECT '
            '(SELECT "baz" FROM "hij") INTERSECT '
            '(SELECT "faz" FROM "lmn")',
            str(query1.intersect(query2).intersect(query3).intersect(query4)),
        )

    def test_intersect_with_order_by(self):
        query1 = Query.from_(self.table1).select(self.table1.foo.as_("a"))
        query2 = Query.from_(self.table2).select(self.table2.bar.as_("a"))

        intersect_query = str((query1.intersect(query2)).orderby(query1.field("a")))

        self.assertEqual(
            '(SELECT "foo" "a" FROM "abc") INTERSECT (SELECT "bar" "a" FROM "efg") ORDER BY "a"',
            intersect_query,
        )

    def test_intersect_with_limit(self):
        query1 = Query.from_(self.table1).select(self.table1.foo.as_("a"))
        query2 = Query.from_(self.table2).select(self.table2.bar.as_("a"))

        intersect_query = str((query1.intersect(query2)).limit(10))

        self.assertEqual(
            '(SELECT "foo" "a" FROM "abc") INTERSECT (SELECT "bar" "a" FROM "efg") LIMIT 10',
            intersect_query,
        )

    def test_intersect_with_offset(self):
        query1 = Query.from_(self.table1).select(self.table1.foo.as_("a"))
        query2 = Query.from_(self.table2).select(self.table2.bar.as_("a"))

        intersect_query = str((query1.intersect(query2)).offset(10))

        self.assertEqual(
            '(SELECT "foo" "a" FROM "abc") INTERSECT (SELECT "bar" "a" FROM "efg") OFFSET 10',
            intersect_query,
        )

    def test_require_equal_number_of_fields_intersect(self):
        query1 = Query.from_(self.table1).select(self.table1.foo)
        query2 = Query.from_(self.table2).select(self.table2.fiz, self.table2.buz)

        with self.assertRaises(SetOperationException):
            str(query1.intersect(query2))

    def test_mysql_query_wraps_intersected_queries(self):
        query1 = MySQLQuery.from_(self.table1).select(self.table1.foo)
        query2 = Query.from_(self.table2).select(self.table2.bar)

        self.assertEqual(
            "(SELECT `foo` FROM `abc`) INTERSECT (SELECT `bar` FROM `efg`)",
            str(query1.intersect(query2)),
        )

    def test_intersect_as_subquery(self):
        abc, efg = Tables("abc", "efg")
        hij = Query.from_(abc).select(abc.t).intersect(Query.from_(efg).select(efg.t))
        q = Query.from_(hij).select(fn.Avg(hij.t))

        self.assertEqual(
            'SELECT AVG("sq0"."t") FROM ((SELECT "t" FROM "abc") INTERSECT (SELECT "t" FROM "efg")) "sq0"',
            str(q),
        )

    def test_intersect_with_no_quote_char(self):
        abc, efg = Tables("abc", "efg")
        hij = Query.from_(abc).select(abc.t).intersect(Query.from_(efg).select(efg.t))
        q = Query.from_(hij).select(fn.Avg(hij.t))

        self.assertEqual(
            "SELECT AVG(sq0.t) FROM ((SELECT t FROM abc) INTERSECT (SELECT t FROM efg)) sq0",
            q.get_sql(quote_char=None),
        )


class MinusTests(unittest.TestCase):
    table1, table2 = Tables("abc", "efg")

    def test_minus(self):
        query1 = Query.from_(self.table1).select(self.table1.foo)
        query2 = Query.from_(self.table2).select(self.table2.bar)

        self.assertEqual('(SELECT "foo" FROM "abc") MINUS (SELECT "bar" FROM "efg")', str(query1.minus(query2)))

        self.assertEqual(
            '(SELECT "foo" FROM "abc") MINUS (SELECT "bar" FROM "efg")',
            str(query1 - query2),
        )

    def test_minus_multiple(self):
        table3, table4 = Tables("hij", "lmn")
        query1 = Query.from_(self.table1).select(self.table1.foo)
        query2 = Query.from_(self.table2).select(self.table2.bar)
        query3 = Query.from_(table3).select(table3.baz)
        query4 = Query.from_(table4).select(table4.faz)

        self.assertEqual(
            '(SELECT "foo" FROM "abc") MINUS '
            '(SELECT "bar" FROM "efg") MINUS '
            '(SELECT "baz" FROM "hij") MINUS '
            '(SELECT "faz" FROM "lmn")',
            str(query1 - query2 - query3 - query4),
        )

        self.assertEqual(
            '(SELECT "foo" FROM "abc") MINUS '
            '(SELECT "bar" FROM "efg") MINUS '
            '(SELECT "baz" FROM "hij") MINUS '
            '(SELECT "faz" FROM "lmn")',
            str(query1.minus(query2).minus(query3).minus(query4)),
        )

    def test_minus_with_order_by(self):
        query1 = Query.from_(self.table1).select(self.table1.foo.as_("a"))
        query2 = Query.from_(self.table2).select(self.table2.bar.as_("a"))

        minus_query = str(query1.minus(query2).orderby(query1.field("a")))

        self.assertEqual(
            '(SELECT "foo" "a" FROM "abc") MINUS (SELECT "bar" "a" FROM "efg") ORDER BY "a"',
            minus_query,
        )

    def test_minus_query_with_order_by_use_minus_query_field(self):
        query1 = Query.from_(self.table1).select(self.table1.foo.as_("a"))
        query2 = Query.from_(self.table2).select(self.table2.bar.as_("a"))

        minus_query = query1.minus(query2)
        minus_query = str(minus_query.orderby(minus_query.field("b")))

        self.assertEqual(
            '(SELECT "foo" "a" FROM "abc") MINUS (SELECT "bar" "a" FROM "efg") ORDER BY "b"',
            minus_query,
        )

    def test_require_equal_number_of_fields(self):
        query1 = Query.from_(self.table1).select(self.table1.foo)
        query2 = Query.from_(self.table2).select(self.table2.fiz, self.table2.buz)

        with self.assertRaises(SetOperationException):
            str(query1.minus(query2))

    def test_mysql_query_wraps_minus_queries(self):
        query1 = MySQLQuery.from_(self.table1).select(self.table1.foo)
        query2 = Query.from_(self.table2).select(self.table2.bar)

        self.assertEqual(
            "(SELECT `foo` FROM `abc`) MINUS (SELECT `bar` FROM `efg`)",
            str(query1 - query2),
        )

    def test_minus_as_subquery(self):
        abc, efg = Tables("abc", "efg")
        hij = Query.from_(abc).select(abc.t).minus(Query.from_(efg).select(efg.t))
        q = Query.from_(hij).select(fn.Avg(hij.t))

        self.assertEqual(
            'SELECT AVG("sq0"."t") FROM ((SELECT "t" FROM "abc") MINUS (SELECT "t" FROM "efg")) "sq0"',
            str(q),
        )

    def test_minus_with_no_quote_char(self):
        abc, efg = Tables("abc", "efg")
        hij = Query.from_(abc).select(abc.t).minus(Query.from_(efg).select(efg.t))
        q = Query.from_(hij).select(fn.Avg(hij.t))

        self.assertEqual(
            "SELECT AVG(sq0.t) FROM ((SELECT t FROM abc) MINUS (SELECT t FROM efg)) sq0",
            q.get_sql(quote_char=None),
        )


class ExceptOfTests(unittest.TestCase):
    table1, table2 = Tables("abc", "efg")

    def test_except(self):
        query1 = Query.from_(self.table1).select(self.table1.foo)
        query2 = Query.from_(self.table2).select(self.table2.bar)

        self.assertEqual('(SELECT "foo" FROM "abc") EXCEPT (SELECT "bar" FROM "efg")', str(query1.except_of(query2)))

    def test_except_multiple(self):
        table3, table4 = Tables("hij", "lmn")
        query1 = Query.from_(self.table1).select(self.table1.foo)
        query2 = Query.from_(self.table2).select(self.table2.bar)
        query3 = Query.from_(table3).select(table3.baz)
        query4 = Query.from_(table4).select(table4.faz)

        self.assertEqual(
            '(SELECT "foo" FROM "abc") EXCEPT '
            '(SELECT "bar" FROM "efg") EXCEPT '
            '(SELECT "baz" FROM "hij") EXCEPT '
            '(SELECT "faz" FROM "lmn")',
            str(query1.except_of(query2).except_of(query3).except_of(query4)),
        )

    def test_except_with_order_by(self):
        query1 = Query.from_(self.table1).select(self.table1.foo.as_("a"))
        query2 = Query.from_(self.table2).select(self.table2.bar.as_("a"))

        except_query = str(query1.except_of(query2).orderby(query1.field("a")))

        self.assertEqual(
            '(SELECT "foo" "a" FROM "abc") EXCEPT (SELECT "bar" "a" FROM "efg") ORDER BY "a"',
            except_query,
        )

    def test_except_query_with_order_by_use_minus_query_field(self):
        query1 = Query.from_(self.table1).select(self.table1.foo.as_("a"))
        query2 = Query.from_(self.table2).select(self.table2.bar.as_("a"))

        except_query = query1.except_of(query2)
        except_query = str(except_query.orderby(except_query.field("b")))

        self.assertEqual(
            '(SELECT "foo" "a" FROM "abc") EXCEPT (SELECT "bar" "a" FROM "efg") ORDER BY "b"',
            except_query,
        )

    def test_require_equal_number_of_fields_except_of(self):
        query1 = Query.from_(self.table1).select(self.table1.foo)
        query2 = Query.from_(self.table2).select(self.table2.fiz, self.table2.buz)

        with self.assertRaises(SetOperationException):
            str(query1.except_of(query2))

    def test_except_as_subquery(self):
        abc, efg = Tables("abc", "efg")
        hij = Query.from_(abc).select(abc.t).except_of(Query.from_(efg).select(efg.t))
        q = Query.from_(hij).select(fn.Avg(hij.t))

        self.assertEqual(
            'SELECT AVG("sq0"."t") FROM ((SELECT "t" FROM "abc") EXCEPT (SELECT "t" FROM "efg")) "sq0"',
            str(q),
        )

    def test_except_with_no_quote_char(self):
        abc, efg = Tables("abc", "efg")
        hij = Query.from_(abc).select(abc.t).except_of(Query.from_(efg).select(efg.t))
        q = Query.from_(hij).select(fn.Avg(hij.t))

        self.assertEqual(
            "SELECT AVG(sq0.t) FROM ((SELECT t FROM abc) EXCEPT (SELECT t FROM efg)) sq0",
            q.get_sql(quote_char=None),
        )



from collections.abc import Sequence
from copy import copy
from functools import reduce
from typing import Any, Optional, Union

from pypika.enums import Dialects, JoinType, ReferenceOption, SetOperation
from pypika.mixins import TupleMixin
from pypika.terms import (
    ArithmeticExpression,
    Criterion,
    EmptyCriterion,
    Field,
    Function,
    Index,
    Node,
    PeriodCriterion,
    Rollup,
    Star,
    Term,
    Tuple,
    ValueWrapper,
)
from pypika.utils import (
    JoinException,
    QueryException,
    RollupException,
    SetOperationException,
    builder,
    format_alias_sql,
    format_quotes,
    ignore_copy,
)

__author__ = "Timothy Heys"
__email__ = "theys@kayak.com"


class Selectable(Node):
    def __init__(self, alias: str) -> None:
        self.alias = alias

    @builder
    def as_(self, alias: str) -> "Selectable":
        self.alias = alias

    def field(self, name: str) -> Field:
        return Field(name, table=self)

    @property
    def star(self) -> Star:
        return Star(self)

    @ignore_copy
    def __getattr__(self, name: str) -> Field:
        return self.field(name)

    @ignore_copy
    def __getitem__(self, name: str) -> Field:
        return self.field(name)

    def get_table_name(self) -> str:
        return self.alias


class AliasedQuery(Selectable):
    def __init__(self, name: str, query: Selectable | None = None) -> None:
        super().__init__(alias=name)
        self.name = name
        self.query = query

    def get_sql(self, **kwargs: Any) -> str:
        if self.query is None:
            return self.name
        return self.query.get_sql(**kwargs)

    def __eq__(self, other: "AliasedQuery") -> bool:
        return isinstance(other, AliasedQuery) and self.name == other.name

    def __hash__(self) -> int:
        return hash(str(self.name))


class Schema:
    def __init__(self, name: str, parent: Optional["Schema"] = None) -> None:
        self._name = name
        self._parent = parent

    def __eq__(self, other: "Schema") -> bool:
        return isinstance(other, Schema) and self._name == other._name and self._parent == other._parent

    def __ne__(self, other: "Schema") -> bool:
        return not self.__eq__(other)

    @ignore_copy
    def __getattr__(self, item: str) -> "Table":
        return Table(item, schema=self)

    def get_sql(self, quote_char: str | None = None, **kwargs: Any) -> str:
        # FIXME escape
        schema_sql = format_quotes(self._name, quote_char)

        if self._parent is not None:
            return f"{self._parent.get_sql(quote_char=quote_char, **kwargs)}.{schema_sql}"

        return schema_sql


class Database(Schema):
    @ignore_copy
    def __getattr__(self, item: str) -> Schema:
        return Schema(item, parent=self)


class Table(Selectable):
    @staticmethod
    def _init_schema(schema: str | list | tuple | Schema | None) -> str | list | tuple | Schema | None:
        # This is a bit complicated in order to support backwards compatibility. It should probably be cleaned up for
        # the next major release. Schema is accepted as a string, list/tuple, Schema instance, or None
        if isinstance(schema, Schema):
            return schema
        if isinstance(schema, (list, tuple)):
            return reduce(lambda obj, s: Schema(s, parent=obj), schema[1:], Schema(schema[0]))
        if schema is not None:
            return Schema(schema)
        return None

    def __init__(
        self,
        name: str,
        schema: Schema | str | None = None,
        alias: str | None = None,
        query_cls: type["Query"] | None = None,
    ) -> None:
        super().__init__(alias)
        self._table_name = name
        self._schema = self._init_schema(schema)
        self._query_cls = query_cls or Query
        self._for = None
        self._for_portion = None
        if not issubclass(self._query_cls, Query):
            raise TypeError("Expected 'query_cls' to be subclass of Query")

    def get_table_name(self) -> str:
        return self.alias or self._table_name

    def get_sql(self, **kwargs: Any) -> str:
        quote_char = kwargs.get("quote_char")
        # FIXME escape
        table_sql = format_quotes(self._table_name, quote_char)

        if self._schema is not None:
            table_sql = f"{self._schema.get_sql(**kwargs)}.{table_sql}"

        if self._for:
            table_sql = f"{table_sql} FOR {self._for.get_sql(**kwargs)}"
        elif self._for_portion:
            table_sql = f"{table_sql} FOR PORTION OF {self._for_portion.get_sql(**kwargs)}"

        return format_alias_sql(table_sql, self.alias, **kwargs)

    @builder
    def for_(self, temporal_criterion: Criterion) -> "Table":
        if self._for:
            raise AttributeError("'Query' object already has attribute for_")
        if self._for_portion:
            raise AttributeError("'Query' object already has attribute for_portion")
        self._for = temporal_criterion

    @builder
    def for_portion(self, period_criterion: PeriodCriterion) -> "Table":
        if self._for_portion:
            raise AttributeError("'Query' object already has attribute for_portion")
        if self._for:
            raise AttributeError("'Query' object already has attribute for_")
        self._for_portion = period_criterion

    def __str__(self) -> str:
        return self.get_sql(quote_char='"')

    def __eq__(self, other) -> bool:
        if not isinstance(other, Table):
            return False

        if self._table_name != other._table_name:
            return False

        if self._schema != other._schema:
            return False

        return not self.alias != other.alias

    def __repr__(self) -> str:
        if self._schema:
            return f"Table('{self._table_name}', schema='{self._schema}')"
        return f"Table('{self._table_name}')"

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(str(self))

    def select(self, *terms: int | float | str | bool | Term | Field) -> "QueryBuilder":
        """
        Perform a SELECT operation on the current table

        :param terms:
            Type:  list[expression]

            A list of terms to select. These can be any type of int, float, str, bool or Term or a Field.

        :return:  QueryBuilder
        """
        return self._query_cls.from_(self).select(*terms)

    def update(self) -> "QueryBuilder":
        """
        Perform an UPDATE operation on the current table

        :return: QueryBuilder
        """
        return self._query_cls.update(self)

    def insert(self, *terms: int | float | str | bool | Term | Field) -> "QueryBuilder":
        """
        Perform an INSERT operation on the current table

        :param terms:
            Type: list[expression]

            A list of terms to select. These can be any type of int, float, str, bool or  any other valid data

        :return: QueryBuilder
        """
        return self._query_cls.into(self).insert(*terms)


class Values(Selectable):
    def __init__(
        self,
        alias: str,
        columns: Sequence[str],
        values: Optional[Sequence[Any]] = None,
        query_cls: Optional[type["Query"]] = None,
    ) -> None:
        super().__init__(alias)
        if isinstance(columns, (list, tuple, set)):
            self.columns = columns
        else:
            self.columns = [columns]

        self._values = []
        if values is not None:
            if not isinstance(values, (list, tuple)):
                raise TypeError("Expected 'values' to be a list or tuple")
            self._validate_terms_and_append(*values)

        self._query_cls = query_cls or Query
        if not issubclass(self._query_cls, Query):
            raise TypeError("Expected 'query_cls' to be subclass of Query")

    def get_sql(self, **kwargs: Any) -> str:
        quote_char = kwargs.get("quote_char")

        return "({values}) AS {alias} ({cols})".format(
            values=self._get_values_sql(),
            alias=format_quotes(self.alias, quote_char),
            cols=self._format_columns(quote_char),
        )

    def __str__(self) -> str:
        return self.get_sql(quote_char='"')

    def __eq__(self, other) -> bool:
        return isinstance(other, Values) and self.alias == other.alias

    def __repr__(self) -> str:
        return "Values('{}', columns=({}))".format(self.alias, ", ".join(["'{}'".format(col) for col in self.columns]))

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(str(self))

    def select(self, *terms: Sequence[Union[int, float, str, bool, Term, Field]]) -> "QueryBuilder":
        """
        Perform a SELECT operation on the current table
        :param terms:
            Type:  list[expression]
            A list of terms to select. These can be any type of int, float, str, bool or Term or a Field.
        :return:  QueryBuilder
        """
        return self._query_cls.from_(self).select(*terms)

    def _validate_terms_and_append(self, *terms: Any) -> None:
        if not isinstance(terms[0], (list, tuple, set)):
            terms = [terms]

        query_cls = self._query_cls
        for values in terms:
            self._values.append(
                [value if isinstance(value, Term) else query_cls.wrap_constant(value) for value in values],
            )

    def add_values(self, *terms: Any) -> None:
        self._validate_terms_and_append(*terms)

    def _format_columns(self, quote_char: str | None) -> str:
        return ",".join(format_quotes(col, quote_char) for col in self.columns)

    def _get_values_sql(self, **kwargs: Any) -> str:
        return "VALUES ({values})".format(
            values="),(".join(self._format_values_row(row, **kwargs) for row in self._values),
        )

    def _format_values_row(self, row: Sequence[Term], **kwargs: Any) -> str:
        return ",".join(term.get_sql(with_alias=True, subquery=True, **kwargs) for term in row)


class ValuesTuple(Values, TupleMixin):
    def __init__(
        self,
        alias: str,
        columns: Sequence[str],
        values: Optional[Sequence[Any]] = None,
        query_cls: Optional[type["Query"]] = None,
        *,
        as_type: str,
        tuple_alias: str = "x",
    ) -> None:
        super().__init__(alias, columns, values, query_cls)
        self.as_type = as_type
        self.tuple_alias = tuple_alias

    def _format_columns(self, quote_char: str | None) -> str:
        return format_quotes(self.tuple_alias, quote_char)

    def _format_values_row(self, row: Sequence[Term], **kwargs: Any) -> str:
        return "({row})::{as_type}".format(row=super()._format_values_row(row, **kwargs), as_type=self.as_type)


def make_tables(*names: tuple[str, str] | str, **kwargs: Any) -> list[Table]:
    """
    Shortcut to create many tables. If `names` param is a tuple, the first
    position will refer to the `_table_name` while the second will be its `alias`.
    Any other data structure will be treated as a whole as the `_table_name`.
    """
    tables = []
    for name in names:
        if isinstance(name, tuple) and len(name) == 2:
            t = Table(
                name=name[0],
                alias=name[1],
                schema=kwargs.get("schema"),
                query_cls=kwargs.get("query_cls"),
            )
        else:
            t = Table(
                name=name,
                schema=kwargs.get("schema"),
                query_cls=kwargs.get("query_cls"),
            )
        tables.append(t)
    return tables


class Column:
    """Represents a column."""

    def __init__(
        self,
        column_name: str,
        column_type: str | None = None,
        nullable: bool | None = None,
        default: Any | Term | None = None,
    ) -> None:
        self.name = column_name
        self.type = column_type
        self.nullable = nullable
        self.default = default if default is None or isinstance(default, Term) else ValueWrapper(default)

    def get_name_sql(self, **kwargs: Any) -> str:
        quote_char = kwargs.get("quote_char")

        column_sql = f"{format_quotes(self.name, quote_char)}"

        return column_sql

    def get_sql(self, **kwargs: Any) -> str:
        column_sql = "{name}{type}{nullable}{default}".format(
            name=self.get_name_sql(**kwargs),
            type=f" {self.type}" if self.type else "",
            nullable=" {}".format("NULL" if self.nullable else "NOT NULL") if self.nullable is not None else "",
            default=" {}".format("DEFAULT " + self.default.get_sql(**kwargs)) if self.default else "",
        )

        return column_sql

    def __str__(self) -> str:
        return self.get_sql(quote_char='"')


def make_columns(*names: tuple[str, str] | str) -> list[Column]:
    """
    Shortcut to create many columns. If `names` param is a tuple, the first
    position will refer to the `name` while the second will be its `type`.
    Any other data structure will be treated as a whole as the `name`.
    """
    columns = []
    for name in names:
        if isinstance(name, tuple) and len(name) == 2:
            column = Column(column_name=name[0], column_type=name[1])
        else:
            column = Column(column_name=name)
        columns.append(column)

    return columns


class PeriodFor:
    def __init__(self, name: str, start_column: str | Column, end_column: str | Column) -> None:
        self.name = name
        self.start_column = start_column if isinstance(start_column, Column) else Column(start_column)
        self.end_column = end_column if isinstance(end_column, Column) else Column(end_column)

    def get_sql(self, **kwargs: Any) -> str:
        quote_char = kwargs.get("quote_char")

        period_for_sql = f"PERIOD FOR {format_quotes(self.name, quote_char)} ({self.start_column.get_name_sql(**kwargs)},{self.end_column.get_name_sql(**kwargs)})"

        return period_for_sql


# for typing in Query's methods
_TableClass = Table


class Query:
    """
    Query is the primary class and entry point in pypika. It is used to build queries iteratively using the builder
    design
    pattern.

    This class is immutable.

    Examples
    --------
    Simple query

    .. code-block:: python

        from pypika import Query, Field
        q = Query.from_('customers').select('*').where(Field("id") == 1)
    """

    @classmethod
    def _builder(cls, **kwargs: Any) -> "QueryBuilder":
        return QueryBuilder(**kwargs)

    @classmethod
    def from_(cls, table: Selectable | str, **kwargs: Any) -> "QueryBuilder":
        """
        Query builder entry point.  Initializes query building and sets the table to select from.  When using this
        function, the query becomes a SELECT query.

        :param table:
            Type: Table or str

            An instance of a Table object or a string table name.

        :return: QueryBuilder
        """
        return cls._builder(**kwargs).from_(table)

    @classmethod
    def create_table(cls, table: str | Table) -> "CreateQueryBuilder":
        """
        Query builder entry point. Initializes query building and sets the table name to be created. When using this
        function, the query becomes a CREATE statement.

        :param table: An instance of a Table object or a string table name.

        :return: CreateQueryBuilder
        """
        return CreateQueryBuilder().create_table(table)

    @classmethod
    def create_index(cls, index: str | Index) -> "CreateIndexBuilder":
        """
        Query builder entry point. Initializes query building and sets the index name to be created. When using this
        function, the query becomes a CREATE statement.
        """
        return CreateIndexBuilder().create_index(index)

    @classmethod
    def drop_database(cls, database: Database | Table) -> "DropQueryBuilder":
        """
        Query builder entry point. Initializes query building and sets the table name to be dropped. When using this
        function, the query becomes a DROP statement.

        :param database: An instance of a Database object or a string database name.

        :return: DropQueryBuilder
        """
        return DropQueryBuilder().drop_database(database)

    @classmethod
    def drop_table(cls, table: str | Table) -> "DropQueryBuilder":
        """
        Query builder entry point. Initializes query building and sets the table name to be dropped. When using this
        function, the query becomes a DROP statement.

        :param table: An instance of a Table object or a string table name.

        :return: DropQueryBuilder
        """
        return DropQueryBuilder().drop_table(table)

    @classmethod
    def drop_user(cls, user: str) -> "DropQueryBuilder":
        """
        Query builder entry point. Initializes query building and sets the table name to be dropped. When using this
        function, the query becomes a DROP statement.

        :param user: String user name.

        :return: DropQueryBuilder
        """
        return DropQueryBuilder().drop_user(user)

    @classmethod
    def drop_view(cls, view: str) -> "DropQueryBuilder":
        """
        Query builder entry point. Initializes query building and sets the table name to be dropped. When using this
        function, the query becomes a DROP statement.

        :param view: String view name.

        :return: DropQueryBuilder
        """
        return DropQueryBuilder().drop_view(view)

    @classmethod
    def drop_index(cls, index: str | Index) -> "DropQueryBuilder":
        """
        Query builder entry point. Initializes query building and sets the index name to be dropped. When using this
        function, the query becomes a DROP statement.
        """
        return DropQueryBuilder().drop_index(index)

    @classmethod
    def into(cls, table: Table | str, **kwargs: Any) -> "QueryBuilder":
        """
        Query builder entry point.  Initializes query building and sets the table to insert into.  When using this
        function, the query becomes an INSERT query.

        :param table:
            Type: Table or str

            An instance of a Table object or a string table name.

        :return QueryBuilder
        """
        return cls._builder(**kwargs).into(table)

    @classmethod
    def with_(cls, table: str | Selectable, name: str, **kwargs: Any) -> "QueryBuilder":
        return cls._builder(**kwargs).with_(table, name)

    @classmethod
    def select(cls, *terms: int | float | str | bool | Term, **kwargs: Any) -> "QueryBuilder":
        """
        Query builder entry point.  Initializes query building without a table and selects fields.  Useful when testing
        SQL functions.

        :param terms:
            Type: list[expression]

            A list of terms to select.  These can be any type of int, float, str, bool, or Term.  They cannot be a Field
            unless the function ``Query.from_`` is called first.

        :return: QueryBuilder
        """
        return cls._builder(**kwargs).select(*terms)

    @classmethod
    def update(cls, table: str | Table, **kwargs) -> "QueryBuilder":
        """
        Query builder entry point.  Initializes query building and sets the table to update.  When using this
        function, the query becomes an UPDATE query.

        :param table:
            Type: Table or str

            An instance of a Table object or a string table name.

        :return: QueryBuilder
        """
        return cls._builder(**kwargs).update(table)

    @classmethod
    def Table(cls, table_name: str, **kwargs) -> _TableClass:
        """
        Convenience method for creating a Table that uses this Query class.

        :param table_name:
            Type: str

            A string table name.

        :return: Table
        """
        kwargs["query_cls"] = cls
        return Table(table_name, **kwargs)

    @classmethod
    def Tables(cls, *names: tuple[str, str] | str, **kwargs: Any) -> list[_TableClass]:
        """
        Convenience method for creating many tables that uses this Query class.
        See ``Query.make_tables`` for details.

        :param names:
            Type: list[str or tuple]

            A list of string table names, or name and alias tuples.

        :return: Table
        """
        kwargs["query_cls"] = cls
        return make_tables(*names, **kwargs)

    @classmethod
    def Values(
        cls,
        alias: str,
        columns: Sequence[str],
        values: Optional[Sequence[Any]] = None,
        **kwargs,
    ) -> Values:
        """
        Convenience method for creating a Table that uses this Query class.
        :param alias:
            Type: str
            A string alias name.
        :param columns:
            Type: list[str]
            A sequence of column names
        :param values:
            Type: list[Any]
            A list (or list of lists) of terms to be in the VALUES statement
        :returns Values
        """
        kwargs["query_cls"] = cls
        return Values(alias, columns, values, **kwargs)


class _SetOperation(Selectable, Term):
    """
    A Query class wrapper for a all set operations, Union DISTINCT or ALL, Intersect, Except or Minus

    Created via the functions `Query.union`,`Query.union_all`,`Query.intersect`, `Query.except_of`,`Query.minus`.

    This class should not be instantiated directly.
    """

    def __init__(
        self,
        base_query: "QueryBuilder",
        set_operation_query: "QueryBuilder",
        set_operation: SetOperation,
        alias: str | None = None,
        wrapper_cls: type[ValueWrapper] = ValueWrapper,
    ):
        super().__init__(alias)
        self.base_query = base_query
        self._set_operation = [(set_operation, set_operation_query)]
        self._orderbys = []

        self._limit = None
        self._offset = None

        self._wrapper_cls = wrapper_cls

    @builder
    def orderby(self, *fields: Field, **kwargs: Any) -> "_SetOperation":
        for field in fields:
            field = (
                Field(field, table=self.base_query._from[0])
                if isinstance(field, str)
                else self.base_query.wrap_constant(field)
            )

            self._orderbys.append((field, kwargs.get("order")))

    @builder
    def limit(self, limit: int) -> "_SetOperation":
        self._limit = limit

    @builder
    def offset(self, offset: int) -> "_SetOperation":
        self._offset = offset

    @builder
    def union(self, other: Selectable) -> "_SetOperation":
        self._set_operation.append((SetOperation.union, other))

    @builder
    def union_all(self, other: Selectable) -> "_SetOperation":
        self._set_operation.append((SetOperation.union_all, other))

    @builder
    def intersect(self, other: Selectable) -> "_SetOperation":
        self._set_operation.append((SetOperation.intersect, other))

    @builder
    def except_of(self, other: Selectable) -> "_SetOperation":
        self._set_operation.append((SetOperation.except_of, other))

    @builder
    def minus(self, other: Selectable) -> "_SetOperation":
        self._set_operation.append((SetOperation.minus, other))

    def __add__(self, other: Selectable) -> "_SetOperation":
        return self.union(other)

    def __mul__(self, other: Selectable) -> "_SetOperation":
        return self.union_all(other)

    def __sub__(self, other: "QueryBuilder") -> "_SetOperation":
        return self.minus(other)

    def __str__(self) -> str:
        return self.get_sql()

    def get_sql(self, with_alias: bool = False, subquery: bool = False, **kwargs: Any) -> str:
        set_operation_template = " {type} {query_string}"

        kwargs.setdefault("dialect", self.base_query.dialect)
        # This initializes the quote char based on the base query, which could be a dialect specific query class
        # This might be overridden if quote_char is set explicitly in kwargs
        kwargs.setdefault("quote_char", self.base_query.QUOTE_CHAR)

        base_querystring = self.base_query.get_sql(subquery=self.base_query.wrap_set_operation_queries, **kwargs)

        querystring = base_querystring
        for set_operation, set_operation_query in self._set_operation:
            set_operation_querystring = set_operation_query.get_sql(
                subquery=self.base_query.wrap_set_operation_queries,
                **kwargs,
            )

            if len(self.base_query._selects) != len(set_operation_query._selects):
                raise SetOperationException(
                    "Queries must have an equal number of select statements in a set operation."
                    f"\n\nMain Query:\n{base_querystring}\n\nSet Operations Query:\n{set_operation_querystring}",
                )

            querystring += set_operation_template.format(
                type=set_operation.value,
                query_string=set_operation_querystring,
            )

        if self._orderbys:
            querystring += self._orderby_sql(**kwargs)

        if self._limit is not None:
            querystring += self._limit_sql()

        if self._offset:
            querystring += self._offset_sql()

        if subquery:
            querystring = "({query})".format(query=querystring, **kwargs)

        if with_alias:
            return format_alias_sql(querystring, self.alias or self._table_name, **kwargs)

        return querystring

    def _orderby_sql(self, quote_char: str | None = None, **kwargs: Any) -> str:
        """
        Produces the ORDER BY part of the query.  This is a list of fields and possibly their directionality, ASC or
        DESC. The clauses are stored in the query under self._orderbys as a list of tuples containing the field and
        directionality (which can be None).

        If an order by field is used in the select clause, determined by a matching , then the ORDER BY clause will use
        the alias, otherwise the field will be rendered as SQL.
        """
        clauses = []
        selected_aliases = {s.alias for s in self.base_query._selects}
        for field, directionality in self._orderbys:
            term = (
                format_quotes(field.alias, quote_char)
                if field.alias and field.alias in selected_aliases
                else field.get_sql(quote_char=quote_char, **kwargs)
            )

            clauses.append(
                f"{term} {directionality.value}" if directionality is not None else term,
            )

        return " ORDER BY {orderby}".format(orderby=",".join(clauses))

    def _offset_sql(self) -> str:
        return f" OFFSET {self._offset}"

    def _limit_sql(self) -> str:
        return f" LIMIT {self._limit}"


class QueryBuilder(Selectable, Term):
    """
    Query Builder is the main class in pypika which stores the state of a query and offers functions which allow the
    state to be branched immutably.
    """

    QUOTE_CHAR = '"'
    SECONDARY_QUOTE_CHAR = "'"
    ALIAS_QUOTE_CHAR = None
    QUERY_ALIAS_QUOTE_CHAR = None
    QUERY_CLS = Query

    def __init__(
        self,
        dialect: Dialects | None = None,
        wrap_set_operation_queries: bool = True,
        wrapper_cls: type[ValueWrapper] = ValueWrapper,
        immutable: bool = True,
        as_keyword: bool = False,
    ):
        super().__init__(None)

        self._from = []
        self._insert_table = None
        self._update_table = None
        self._delete_from = False
        self._replace = False

        self._with = []
        self._selects = []
        self._force_indexes = []
        self._use_indexes = []
        self._columns = []
        self._values = []
        self._distinct = False
        self._ignore = False

        self._for_update = False

        self._wheres = None
        self._prewheres = None
        self._groupbys = []
        self._with_totals = False
        self._havings = None
        self._orderbys = []
        self._joins = []
        self._unions = []
        self._using = []

        self._limit = None
        self._offset = None

        self._updates = []

        self._select_star = False
        self._select_star_tables = set()
        self._mysql_rollup = False
        self._select_into = False

        self._subquery_count = 0
        self._foreign_table = False

        self.dialect = dialect
        self.as_keyword = as_keyword
        self.wrap_set_operation_queries = wrap_set_operation_queries

        self._wrapper_cls = wrapper_cls

        self.immutable = immutable

    def __copy__(self) -> "QueryBuilder":
        newone = type(self).__new__(type(self))
        newone.__dict__.update(self.__dict__)
        newone._select_star_tables = copy(self._select_star_tables)
        newone._from = copy(self._from)
        newone._with = copy(self._with)
        newone._selects = copy(self._selects)
        newone._columns = copy(self._columns)
        newone._values = copy(self._values)
        newone._groupbys = copy(self._groupbys)
        newone._orderbys = copy(self._orderbys)
        newone._joins = copy(self._joins)
        newone._unions = copy(self._unions)
        newone._updates = copy(self._updates)
        newone._force_indexes = copy(self._force_indexes)
        newone._use_indexes = copy(self._use_indexes)
        return newone

    @builder
    def from_(self, selectable: Selectable | Query | str) -> "QueryBuilder":
        """
        Adds a table to the query. This function can only be called once and will raise an AttributeError if called a
        second time.

        :param selectable:
            Type: ``Table``, ``Query``, or ``str``

            When a ``str`` is passed, a table with the name matching the ``str`` value is used.

        :returns
            A copy of the query with the table added.
        """

        self._from.append(Table(selectable) if isinstance(selectable, str) else selectable)

        if isinstance(selectable, (QueryBuilder, _SetOperation)) and selectable.alias is None:
            sub_query_count = selectable._subquery_count if isinstance(selectable, QueryBuilder) else 0

            sub_query_count = max(self._subquery_count, sub_query_count)
            selectable.alias = "sq%d" % sub_query_count
            self._subquery_count = sub_query_count + 1

    @builder
    def replace_table(self, current_table: Table | None, new_table: Table | None) -> "QueryBuilder":
        """
        Replaces all occurrences of the specified table with the new table. Useful when reusing fields across
        queries.

        :param current_table:
            The table instance to be replaces.
        :param new_table:
            The table instance to replace with.
        :return:
            A copy of the query with the tables replaced.
        """
        self._from = [new_table if table == current_table else table for table in self._from]
        self._insert_table = new_table if self._insert_table == current_table else self._insert_table
        self._update_table = new_table if self._update_table == current_table else self._update_table

        self._with = [alias_query.replace_table(current_table, new_table) for alias_query in self._with]
        self._selects = [select.replace_table(current_table, new_table) for select in self._selects]
        self._columns = [column.replace_table(current_table, new_table) for column in self._columns]
        self._values = [
            [value.replace_table(current_table, new_table) for value in value_list] for value_list in self._values
        ]

        self._wheres = self._wheres.replace_table(current_table, new_table) if self._wheres else None
        self._prewheres = self._prewheres.replace_table(current_table, new_table) if self._prewheres else None
        self._groupbys = [groupby.replace_table(current_table, new_table) for groupby in self._groupbys]
        self._havings = self._havings.replace_table(current_table, new_table) if self._havings else None
        self._orderbys = [
            (orderby[0].replace_table(current_table, new_table), orderby[1]) for orderby in self._orderbys
        ]
        self._joins = [join.replace_table(current_table, new_table) for join in self._joins]

        if current_table in self._select_star_tables:
            self._select_star_tables.remove(current_table)
            self._select_star_tables.add(new_table)

    @builder
    def with_(self, selectable: Selectable, name: str) -> "QueryBuilder":
        t = AliasedQuery(name, selectable)
        self._with.append(t)

    @builder
    def into(self, table: str | Table) -> "QueryBuilder":
        if self._insert_table is not None:
            raise AttributeError("'Query' object has no attribute '%s'" % "into")

        if self._selects:
            self._select_into = True

        self._insert_table = table if isinstance(table, Table) else Table(table)

    @builder
    def select(self, *terms: Any) -> "QueryBuilder":
        for term in terms:
            if isinstance(term, Field):
                self._select_field(term)
            elif isinstance(term, str):
                self._select_field_str(term)
            elif isinstance(term, (Function, ArithmeticExpression)):
                self._select_other(term)
            else:
                self._select_other(self.wrap_constant(term, wrapper_cls=self._wrapper_cls))

    @builder
    def delete(self) -> "QueryBuilder":
        if self._delete_from or self._selects or self._update_table:
            raise AttributeError("'Query' object has no attribute '%s'" % "delete")

        self._delete_from = True

    @builder
    def update(self, table: str | Table) -> "QueryBuilder":
        if self._update_table is not None or self._selects or self._delete_from:
            raise AttributeError("'Query' object has no attribute '%s'" % "update")

        self._update_table = table if isinstance(table, Table) else Table(table)

    @builder
    def columns(self, *terms: Any) -> "QueryBuilder":
        if self._insert_table is None:
            raise AttributeError("'Query' object has no attribute '%s'" % "insert")

        if terms and isinstance(terms[0], (list, tuple)):
            terms = terms[0]

        for term in terms:
            if isinstance(term, str):
                term = Field(term, table=self._insert_table)
            self._columns.append(term)

    @builder
    def insert(self, *terms: Any) -> "QueryBuilder":
        self._apply_terms(*terms)
        self._replace = False

    @builder
    def replace(self, *terms: Any) -> "QueryBuilder":
        self._apply_terms(*terms)
        self._replace = True

    @builder
    def force_index(self, term: str | Index, *terms: str | Index) -> "QueryBuilder":
        for t in (term, *terms):
            if isinstance(t, Index):
                self._force_indexes.append(t)
            elif isinstance(t, str):
                self._force_indexes.append(Index(t))

    @builder
    def use_index(self, term: str | Index, *terms: str | Index) -> "QueryBuilder":
        for t in (term, *terms):
            if isinstance(t, Index):
                self._use_indexes.append(t)
            elif isinstance(t, str):
                self._use_indexes.append(Index(t))

    @builder
    def distinct(self) -> "QueryBuilder":
        self._distinct = True

    @builder
    def for_update(self) -> "QueryBuilder":
        self._for_update = True

    @builder
    def ignore(self) -> "QueryBuilder":
        self._ignore = True

    @builder
    def prewhere(self, criterion: Criterion) -> "QueryBuilder":
        if not self._validate_table(criterion):
            self._foreign_table = True

        if self._prewheres:
            self._prewheres &= criterion
        else:
            self._prewheres = criterion

    @builder
    def where(self, criterion: Term | EmptyCriterion) -> "QueryBuilder":
        if isinstance(criterion, EmptyCriterion):
            return

        if not self._validate_table(criterion):
            self._foreign_table = True

        if self._wheres:
            self._wheres &= criterion
        else:
            self._wheres = criterion

    @builder
    def having(self, criterion: Term | EmptyCriterion) -> "QueryBuilder":
        if isinstance(criterion, EmptyCriterion):
            return

        if self._havings:
            self._havings &= criterion
        else:
            self._havings = criterion

    @builder
    def groupby(self, *terms: str | int | Term) -> "QueryBuilder":
        for term in terms:
            if isinstance(term, str):
                term = Field(term, table=self._from[0])
            elif isinstance(term, int):
                term = Field(str(term), table=self._from[0]).wrap_constant(term)

            self._groupbys.append(term)

    @builder
    def with_totals(self) -> "QueryBuilder":
        self._with_totals = True

    @builder
    def rollup(self, *terms: list | tuple | set | Term, **kwargs: Any) -> "QueryBuilder":
        for_mysql = kwargs.get("vendor") == "mysql"

        if self._mysql_rollup:
            raise AttributeError("'Query' object has no attribute '%s'" % "rollup")

        terms = [Tuple(*term) if isinstance(term, (list, tuple, set)) else term for term in terms]

        if for_mysql:
            # MySQL rolls up all of the dimensions always
            if not terms and not self._groupbys:
                raise RollupException(
                    "At least one group is required. Call Query.groupby(term) or pass as parameter to rollup.",
                )

            self._mysql_rollup = True
            self._groupbys += terms

        elif len(self._groupbys) > 0 and isinstance(self._groupbys[-1], Rollup):
            # If a rollup was added last, then append the new terms to the previous rollup
            self._groupbys[-1].args += terms

        else:
            self._groupbys.append(Rollup(*terms))

    @builder
    def orderby(self, *fields: Any, **kwargs: Any) -> "QueryBuilder":
        for field in fields:
            field = Field(field, table=self._from[0]) if isinstance(field, str) else self.wrap_constant(field)

            self._orderbys.append((field, kwargs.get("order")))

    @builder
    def join(
        self,
        item: Union[Table, "QueryBuilder", AliasedQuery, Selectable],
        how: JoinType = JoinType.inner,
    ) -> "Joiner":
        if isinstance(item, Table):
            return Joiner(self, item, how, type_label="table")

        elif isinstance(item, QueryBuilder):
            if item.alias is None:
                self._tag_subquery(item)
            return Joiner(self, item, how, type_label="subquery")

        elif isinstance(item, AliasedQuery):
            return Joiner(self, item, how, type_label="table")

        elif isinstance(item, Selectable):
            return Joiner(self, item, how, type_label="subquery")

        raise ValueError("Cannot join on type '%s'" % type(item))

    def inner_join(self, item: Union[Table, "QueryBuilder", AliasedQuery]) -> "Joiner":
        return self.join(item, JoinType.inner)

    def left_join(self, item: Union[Table, "QueryBuilder", AliasedQuery]) -> "Joiner":
        return self.join(item, JoinType.left)

    def left_outer_join(self, item: Union[Table, "QueryBuilder", AliasedQuery]) -> "Joiner":
        return self.join(item, JoinType.left_outer)

    def right_join(self, item: Union[Table, "QueryBuilder", AliasedQuery]) -> "Joiner":
        return self.join(item, JoinType.right)

    def right_outer_join(self, item: Union[Table, "QueryBuilder", AliasedQuery]) -> "Joiner":
        return self.join(item, JoinType.right_outer)

    def outer_join(self, item: Union[Table, "QueryBuilder", AliasedQuery]) -> "Joiner":
        return self.join(item, JoinType.outer)

    def full_outer_join(self, item: Union[Table, "QueryBuilder", AliasedQuery]) -> "Joiner":
        return self.join(item, JoinType.full_outer)

    def cross_join(self, item: Union[Table, "QueryBuilder", AliasedQuery]) -> "Joiner":
        return self.join(item, JoinType.cross)

    def hash_join(self, item: Union[Table, "QueryBuilder", AliasedQuery]) -> "Joiner":
        return self.join(item, JoinType.hash)

    @builder
    def limit(self, limit: int) -> "QueryBuilder":
        self._limit = limit

    @builder
    def offset(self, offset: int) -> "QueryBuilder":
        self._offset = offset

    @builder
    def union(self, other: "QueryBuilder") -> _SetOperation:
        return _SetOperation(self, other, SetOperation.union, wrapper_cls=self._wrapper_cls)

    @builder
    def union_all(self, other: "QueryBuilder") -> _SetOperation:
        return _SetOperation(self, other, SetOperation.union_all, wrapper_cls=self._wrapper_cls)

    @builder
    def intersect(self, other: "QueryBuilder") -> _SetOperation:
        return _SetOperation(self, other, SetOperation.intersect, wrapper_cls=self._wrapper_cls)

    @builder
    def except_of(self, other: "QueryBuilder") -> _SetOperation:
        return _SetOperation(self, other, SetOperation.except_of, wrapper_cls=self._wrapper_cls)

    @builder
    def minus(self, other: "QueryBuilder") -> _SetOperation:
        return _SetOperation(self, other, SetOperation.minus, wrapper_cls=self._wrapper_cls)

    @builder
    def set(self, field: Field | str, value: Any) -> "QueryBuilder":
        field = Field(field) if not isinstance(field, Field) else field
        self._updates.append((field, self._wrapper_cls(value)))

    def __add__(self, other: "QueryBuilder") -> _SetOperation:
        return self.union(other)

    def __mul__(self, other: "QueryBuilder") -> _SetOperation:
        return self.union_all(other)

    def __sub__(self, other: "QueryBuilder") -> _SetOperation:
        return self.minus(other)

    @builder
    def slice(self, slice: slice) -> "QueryBuilder":
        self._offset = slice.start
        self._limit = slice.stop

    def __getitem__(self, item: Any) -> Union["QueryBuilder", Field]:
        if not isinstance(item, slice):
            return super().__getitem__(item)
        return self.slice(item)

    @staticmethod
    def _list_aliases(field_set: Sequence[Field], quote_char: str | None = None) -> list[str]:
        return [field.alias or field.get_sql(quote_char=quote_char) for field in field_set]

    def _select_field_str(self, term: str) -> None:
        if len(self._from) == 0:
            raise QueryException(f"Cannot select {term}, no FROM table specified.")

        if term == "*":
            self._select_star = True
            self._selects = [Star()]
            return

        self._select_field(Field(term, table=self._from[0]))

    def _select_field(self, term: Field) -> None:
        if self._select_star:
            # Do not add select terms after a star is selected
            return

        if term.table in self._select_star_tables:
            # Do not add select terms for table after a table star is selected
            return

        if isinstance(term, Star):
            self._selects = [
                select for select in self._selects if not hasattr(select, "table") or term.table != select.table
            ]
            self._select_star_tables.add(term.table)

        self._selects.append(term)

    def _select_other(self, function: Function) -> None:
        self._selects.append(function)

    def fields_(self) -> list[Field]:
        # Don't return anything here. Subqueries have their own fields.
        return []

    def do_join(self, join: "Join") -> None:
        base_tables = [*self._from, self._update_table, *self._with]
        join.validate(base_tables, self._joins)

        table_in_query = any(isinstance(clause, Table) and join.item in base_tables for clause in base_tables)
        if isinstance(join.item, Table) and join.item.alias is None and table_in_query:
            # On the odd chance that we join the same table as the FROM table and don't set an alias
            # FIXME only works once
            join.item.alias = join.item._table_name + "2"

        self._joins.append(join)

    def is_joined(self, table: Table) -> bool:
        return any(table == join.item for join in self._joins)

    def _validate_table(self, term: Term) -> bool:
        """
        Returns False if the term references a table not already part of the
        FROM clause or JOINS and True otherwise.
        """
        base_tables = [*self._from, self._update_table]

        for field in term.fields_():
            table_in_base_tables = field.table in base_tables
            table_in_joins = field.table in [join.item for join in self._joins]
            if all(
                [
                    field.table is not None,
                    not table_in_base_tables,
                    not table_in_joins,
                    field.table != self._update_table,
                ],
            ):
                return False
        return True

    def _tag_subquery(self, subquery: "QueryBuilder") -> None:
        subquery.alias = "sq%d" % self._subquery_count
        self._subquery_count += 1

    def _apply_terms(self, *terms: Any) -> None:
        """
        Handy function for INSERT and REPLACE statements in order to check if
        terms are introduced and how append them to `self._values`
        """
        if self._insert_table is None:
            raise AttributeError("'Query' object has no attribute '%s'" % "insert")

        if not terms:
            return

        if not isinstance(terms[0], (list, tuple, set)):
            terms = [terms]

        for values in terms:
            self._values.append([value if isinstance(value, Term) else self.wrap_constant(value) for value in values])

    def __str__(self) -> str:
        return self.get_sql(dialect=self.dialect)

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: "QueryBuilder") -> bool:
        if not isinstance(other, QueryBuilder):
            return False

        return self.alias == other.alias

    def __ne__(self, other: "QueryBuilder") -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.alias) + sum(hash(clause) for clause in self._from)

    def _set_kwargs_defaults(self, kwargs: dict) -> None:
        kwargs.setdefault("quote_char", self.QUOTE_CHAR)
        kwargs.setdefault("secondary_quote_char", self.SECONDARY_QUOTE_CHAR)
        kwargs.setdefault("alias_quote_char", self.ALIAS_QUOTE_CHAR)
        kwargs.setdefault("as_keyword", self.as_keyword)
        kwargs.setdefault("dialect", self.dialect)

    def get_sql(self, with_alias: bool = False, subquery: bool = False, **kwargs: Any) -> str:
        self._set_kwargs_defaults(kwargs)
        if not (self._selects or self._insert_table or self._delete_from or self._update_table):
            return ""
        if self._insert_table and not (self._selects or self._values):
            return ""
        if self._update_table and not self._updates:
            return ""

        has_joins = bool(self._joins)
        has_multiple_from_clauses = len(self._from) > 1
        has_subquery_from_clause = len(self._from) > 0 and isinstance(self._from[0], QueryBuilder)
        has_reference_to_foreign_table = self._foreign_table
        has_update_from = self._update_table and self._from

        kwargs["with_namespace"] = any(
            [
                has_joins,
                has_multiple_from_clauses,
                has_subquery_from_clause,
                has_reference_to_foreign_table,
                has_update_from,
            ],
        )

        if self._update_table:
            querystring = self._with_sql(**kwargs) if self._with else ""

            querystring += self._update_sql(**kwargs)

            if self._joins:
                querystring += " " + " ".join(join.get_sql(**kwargs) for join in self._joins)

            querystring += self._set_sql(**kwargs)

            if self._from:
                querystring += self._from_sql(**kwargs)

            if self._wheres:
                querystring += self._where_sql(**kwargs)

            if self._limit is not None:
                querystring += self._limit_sql()

            return querystring

        if self._delete_from:
            querystring = self._delete_sql(**kwargs)

        elif not self._select_into and self._insert_table:
            querystring = self._with_sql(**kwargs) if self._with else ""

            if self._replace:
                querystring += self._replace_sql(**kwargs)
            else:
                querystring += self._insert_sql(**kwargs)

            if self._columns:
                querystring += self._columns_sql(**kwargs)

            if self._values:
                querystring += self._values_sql(**kwargs)
                return querystring
            else:
                querystring += " " + self._select_sql(**kwargs)

        else:
            querystring = self._with_sql(**kwargs) if self._with else ""

            querystring += self._select_sql(**kwargs)

            if self._insert_table:
                querystring += self._into_sql(**kwargs)

        if self._from:
            querystring += self._from_sql(**kwargs)

        if self._using:
            querystring += self._using_sql(**kwargs)

        if self._force_indexes:
            querystring += self._force_index_sql(**kwargs)

        if self._use_indexes:
            querystring += self._use_index_sql(**kwargs)

        if self._joins:
            querystring += " " + " ".join(join.get_sql(**kwargs) for join in self._joins)

        if self._prewheres:
            querystring += self._prewhere_sql(**kwargs)

        if self._wheres:
            querystring += self._where_sql(**kwargs)

        if self._groupbys:
            querystring += self._group_sql(**kwargs)
            if self._mysql_rollup:
                querystring += self._rollup_sql()

        if self._havings:
            querystring += self._having_sql(**kwargs)

        if self._orderbys:
            querystring += self._orderby_sql(**kwargs)

        querystring = self._apply_pagination(querystring, **kwargs)

        if self._for_update:
            querystring += self._for_update_sql(**kwargs)

        if subquery:
            querystring = f"({querystring})"

        if with_alias:
            kwargs['alias_quote_char'] = (
                self.ALIAS_QUOTE_CHAR if self.QUERY_ALIAS_QUOTE_CHAR is None else self.QUERY_ALIAS_QUOTE_CHAR
            )
            return format_alias_sql(querystring, self.alias, **kwargs)

        return querystring

    def _apply_pagination(self, querystring: str, **kwargs) -> str:
        if self._limit is not None:
            querystring += self._limit_sql()

        if self._offset:
            querystring += self._offset_sql()

        return querystring

    def _with_sql(self, **kwargs: Any) -> str:
        return "WITH " + ",".join(
            clause.name + " AS (" + clause.get_sql(subquery=False, with_alias=False, **kwargs) + ") "
            for clause in self._with
        )

    def _distinct_sql(self, **kwargs: Any) -> str:
        distinct = "DISTINCT " if self._distinct else ""

        return distinct

    def _for_update_sql(self, **kwargs) -> str:
        for_update = " FOR UPDATE" if self._for_update else ""

        return for_update

    def _select_sql(self, **kwargs: Any) -> str:
        return "SELECT {distinct}{select}".format(
            distinct=self._distinct_sql(**kwargs),
            select=",".join(term.get_sql(with_alias=True, subquery=True, **kwargs) for term in self._selects),
        )

    def _insert_sql(self, **kwargs: Any) -> str:
        return "INSERT {ignore}INTO {table}".format(
            table=self._insert_table.get_sql(**kwargs),
            ignore="IGNORE " if self._ignore else "",
        )

    def _replace_sql(self, **kwargs: Any) -> str:
        return f"REPLACE INTO {self._insert_table.get_sql(**kwargs)}"

    @staticmethod
    def _delete_sql(**kwargs: Any) -> str:
        return "DELETE"

    def _update_sql(self, **kwargs: Any) -> str:
        return f"UPDATE {self._update_table.get_sql(**kwargs)}"

    def _columns_sql(self, with_namespace: bool = False, **kwargs: Any) -> str:
        """
        SQL for Columns clause for INSERT queries
        :param with_namespace:
            Remove from kwargs, never format the column terms with namespaces since only one table can be inserted into
        """
        return " ({columns})".format(
            columns=",".join(term.get_sql(with_namespace=False, **kwargs) for term in self._columns),
        )

    def _values_sql(self, **kwargs: Any) -> str:
        return " VALUES ({values})".format(
            values="),(".join(
                ",".join(term.get_sql(with_alias=True, subquery=True, **kwargs) for term in row) for row in self._values
            ),
        )

    def _into_sql(self, **kwargs: Any) -> str:
        return f" INTO {self._insert_table.get_sql(with_alias=False, **kwargs)}"

    def _from_sql(self, with_namespace: bool = False, **kwargs: Any) -> str:
        return " FROM {selectable}".format(
            selectable=",".join(clause.get_sql(subquery=True, with_alias=True, **kwargs) for clause in self._from),
        )

    def _using_sql(self, with_namespace: bool = False, **kwargs: Any) -> str:
        return " USING {selectable}".format(
            selectable=",".join(clause.get_sql(subquery=True, with_alias=True, **kwargs) for clause in self._using),
        )

    def _force_index_sql(self, **kwargs: Any) -> str:
        return " FORCE INDEX ({indexes})".format(
            indexes=",".join(index.get_sql(**kwargs) for index in self._force_indexes),
        )

    def _use_index_sql(self, **kwargs: Any) -> str:
        return " USE INDEX ({indexes})".format(
            indexes=",".join(index.get_sql(**kwargs) for index in self._use_indexes),
        )

    def _prewhere_sql(self, quote_char: str | None = None, **kwargs: Any) -> str:
        return f" PREWHERE {self._prewheres.get_sql(quote_char=quote_char, subquery=True, **kwargs)}"

    def _where_sql(self, quote_char: str | None = None, **kwargs: Any) -> str:
        return f" WHERE {self._wheres.get_sql(quote_char=quote_char, subquery=True, **kwargs)}"

    def _group_sql(
        self,
        quote_char: str | None = None,
        alias_quote_char: str | None = None,
        groupby_alias: bool = True,
        **kwargs: Any,
    ) -> str:
        """
        Produces the GROUP BY part of the query.  This is a list of fields. The clauses are stored in the query under
        self._groupbys as a list fields.

        If an groupby field is used in the select clause,
        determined by a matching alias, and the groupby_alias is set True
        then the GROUP BY clause will use the alias,
        otherwise the entire field will be rendered as SQL.
        """
        clauses = []
        selected_aliases = {s.alias for s in self._selects}
        for field in self._groupbys:
            if groupby_alias and field.alias and field.alias in selected_aliases:
                clauses.append(format_quotes(field.alias, alias_quote_char or quote_char))
            else:
                clauses.append(field.get_sql(quote_char=quote_char, alias_quote_char=alias_quote_char, **kwargs))

        sql = " GROUP BY {groupby}".format(groupby=",".join(clauses))

        if self._with_totals:
            return sql + " WITH TOTALS"

        return sql

    def _orderby_sql(
        self,
        quote_char: str | None = None,
        alias_quote_char: str | None = None,
        orderby_alias: bool = True,
        **kwargs: Any,
    ) -> str:
        """
        Produces the ORDER BY part of the query.  This is a list of fields and possibly their directionality, ASC or
        DESC. The clauses are stored in the query under self._orderbys as a list of tuples containing the field and
        directionality (which can be None).

        If an order by field is used in the select clause,
        determined by a matching, and the orderby_alias
        is set True then the ORDER BY clause will use
        the alias, otherwise the field will be rendered as SQL.
        """
        clauses = []
        selected_aliases = {s.alias for s in self._selects}
        for field, directionality in self._orderbys:
            term = (
                format_quotes(field.alias, alias_quote_char or quote_char)
                if orderby_alias and field.alias and field.alias in selected_aliases
                else field.get_sql(quote_char=quote_char, alias_quote_char=alias_quote_char, **kwargs)
            )

            clauses.append(
                f"{term} {directionality.value}" if directionality is not None else term,
            )

        return " ORDER BY {orderby}".format(orderby=",".join(clauses))

    def _rollup_sql(self) -> str:
        return " WITH ROLLUP"

    def _having_sql(self, quote_char: str | None = None, **kwargs: Any) -> str:
        return f" HAVING {self._havings.get_sql(quote_char=quote_char, **kwargs)}"

    def _offset_sql(self) -> str:
        return f" OFFSET {self._offset}"

    def _limit_sql(self) -> str:
        return f" LIMIT {self._limit}"

    def _set_sql(self, **kwargs: Any) -> str:
        return " SET {set}".format(
            set=",".join(
                f"{field.get_sql(**dict(kwargs, with_namespace=False))}={value.get_sql(**kwargs)}"
                for field, value in self._updates
            ),
        )

    def pipe(self, func, *args, **kwargs):
        """Call a function on the current object and return the result.

        Example usage:

        .. code-block:: python

            from pypika import Query, functions as fn
            from pypika.queries import QueryBuilder

            def rows_by_group(query: QueryBuilder, *groups) -> QueryBuilder:
                return (
                    query
                    .select(*groups, fn.Count("*").as_("n_rows"))
                    .groupby(*groups)
                )

            base_query = Query.from_("table")

            col1_agg = base_query.pipe(rows_by_group, "col1")
            col2_agg = base_query.pipe(rows_by_group, "col2")
            col1_col2_agg = base_query.pipe(rows_by_group, "col1", "col2")

        Makes chaining functions together easier, especially when the functions are
        defined elsewhere. For example, you could define a function that filters
        rows by a date range and then group by a set of columns:


        .. code-block:: python

            from datetime import datetime, timedelta

            from pypika import Field

            def days_since(query: QueryBuilder, n_days: int) -> QueryBuilder:
                return (
                    query
                    .where("date" > fn.Date(datetime.now().date() - timedelta(days=n_days)))
                )

            (
                base_query
                .pipe(days_since, n_days=7)
                .pipe(rows_by_group, "col1", "col2")
            )
        """
        return func(self, *args, **kwargs)


class Joiner:
    def __init__(
        self,
        query: QueryBuilder,
        item: Union[Table, "QueryBuilder", AliasedQuery],
        how: JoinType,
        type_label: str,
    ) -> None:
        self.query = query
        self.item = item
        self.how = how
        self.type_label = type_label

    def on(self, criterion: Criterion | None, collate: str | None = None) -> QueryBuilder:
        if criterion is None:
            raise JoinException(
                f"Parameter 'criterion' is required for a {self.type_label} JOIN but was not supplied.",
            )

        self.query.do_join(JoinOn(self.item, self.how, criterion, collate))
        return self.query

    def on_field(self, *fields: Any) -> QueryBuilder:
        if not fields:
            raise JoinException(
                f"Parameter 'fields' is required for a {self.type_label} JOIN but was not supplied.",
            )

        criterion = None
        for field in fields:
            constituent = Field(field, table=self.query._from[0]) == Field(field, table=self.item)
            criterion = constituent if criterion is None else criterion & constituent

        self.query.do_join(JoinOn(self.item, self.how, criterion))
        return self.query

    def using(self, *fields: Any) -> QueryBuilder:
        if not fields:
            raise JoinException("Parameter 'fields' is required when joining with a using clause but was not supplied.")

        self.query.do_join(JoinUsing(self.item, self.how, [Field(field) for field in fields]))
        return self.query

    def cross(self) -> QueryBuilder:
        """Return cross join"""
        self.query.do_join(Join(self.item, JoinType.cross))

        return self.query


class Join:
    def __init__(self, item: Term, how: JoinType) -> None:
        self.item = item
        self.how = how

    def get_sql(self, **kwargs: Any) -> str:
        sql = f"JOIN {self.item.get_sql(subquery=True, with_alias=True, **kwargs)}"

        if self.how.value:
            return f"{self.how.value} {sql}"
        return sql

    def validate(self, _from: Sequence[Table], _joins: Sequence[Table]) -> None:
        pass

    @builder
    def replace_table(self, current_table: Table | None, new_table: Table | None) -> "Join":
        """
        Replaces all occurrences of the specified table with the new table. Useful when reusing
        fields across queries.

        :param current_table:
            The table to be replaced.
        :param new_table:
            The table to replace with.
        :return:
            A copy of the join with the tables replaced.
        """
        self.item = self.item.replace_table(current_table, new_table)


class JoinOn(Join):
    def __init__(self, item: Term, how: JoinType, criteria: QueryBuilder, collate: str | None = None) -> None:
        super().__init__(item, how)
        self.criterion = criteria
        self.collate = collate

    def get_sql(self, **kwargs: Any) -> str:
        join_sql = super().get_sql(**kwargs)
        return "{join} ON {criterion}{collate}".format(
            join=join_sql,
            criterion=self.criterion.get_sql(subquery=True, **kwargs),
            collate=f" COLLATE {self.collate}" if self.collate else "",
        )

    def validate(self, _from: Sequence[Table], _joins: Sequence[Table]) -> None:
        criterion_tables = {f.table for f in self.criterion.fields_()}
        available_tables = set(_from) | {join.item for join in _joins} | {self.item}
        missing_tables = criterion_tables - available_tables
        if missing_tables:
            raise JoinException(
                "Invalid join criterion. One field is required from the joined item and "
                "another from the selected table or an existing join.  Found [{tables}]".format(
                    tables=", ".join(map(str, missing_tables)),
                ),
            )

    @builder
    def replace_table(self, current_table: Table | None, new_table: Table | None) -> "JoinOn":
        """
        Replaces all occurrences of the specified table with the new table. Useful when reusing
        fields across queries.

        :param current_table:
            The table to be replaced.
        :param new_table:
            The table to replace with.
        :return:
            A copy of the join with the tables replaced.
        """
        self.item = new_table if self.item == current_table else self.item
        self.criterion = self.criterion.replace_table(current_table, new_table)


class JoinUsing(Join):
    def __init__(self, item: Term, how: JoinType, fields: Sequence[Field]) -> None:
        super().__init__(item, how)
        self.fields = fields

    def get_sql(self, **kwargs: Any) -> str:
        join_sql = super().get_sql(**kwargs)
        return "{join} USING ({fields})".format(
            join=join_sql,
            fields=",".join(field.get_sql(**kwargs) for field in self.fields),
        )

    def validate(self, _from: Sequence[Table], _joins: Sequence[Table]) -> None:
        pass

    @builder
    def replace_table(self, current_table: Table | None, new_table: Table | None) -> "JoinUsing":
        """
        Replaces all occurrences of the specified table with the new table. Useful when reusing
        fields across queries.

        :param current_table:
            The table to be replaced.
        :param new_table:
            The table to replace with.
        :return:
            A copy of the join with the tables replaced.
        """
        self.item = new_table if self.item == current_table else self.item
        self.fields = [field.replace_table(current_table, new_table) for field in self.fields]


class CreateQueryBuilder:
    """
    Query builder used to build CREATE queries.
    """

    QUOTE_CHAR = '"'
    SECONDARY_QUOTE_CHAR = "'"
    ALIAS_QUOTE_CHAR = None
    QUERY_CLS = Query

    def __init__(self, dialect: Dialects | None = None) -> None:
        self._create_table = None
        self._temporary = False
        self._unlogged = False
        self._as_select = None
        self._columns = []
        self._period_fors = []
        self._with_system_versioning = False
        self._primary_key = None
        self._uniques = []
        self._if_not_exists = False
        self.dialect = dialect
        self._foreign_key = None
        self._foreign_key_reference_table = None
        self._foreign_key_reference = None
        self._foreign_key_on_update: ReferenceOption = None
        self._foreign_key_on_delete: ReferenceOption = None

    def _set_kwargs_defaults(self, kwargs: dict) -> None:
        kwargs.setdefault("quote_char", self.QUOTE_CHAR)
        kwargs.setdefault("secondary_quote_char", self.SECONDARY_QUOTE_CHAR)
        kwargs.setdefault("dialect", self.dialect)

    @builder
    def create_table(self, table: Table | str) -> "CreateQueryBuilder":
        """
        Creates the table.

        :param table:
            An instance of a Table object or a string table name.

        :raises AttributeError:
            If the table is already created.

        :return:
            CreateQueryBuilder.
        """
        if self._create_table:
            raise AttributeError("'Query' object already has attribute create_table")

        self._create_table = table if isinstance(table, Table) else Table(table)

    @builder
    def temporary(self) -> "CreateQueryBuilder":
        """
        Makes the table temporary.

        :return:
            CreateQueryBuilder.
        """
        self._temporary = True

    @builder
    def unlogged(self) -> "CreateQueryBuilder":
        """
        Makes the table unlogged.

        :return:
            CreateQueryBuilder.
        """
        self._unlogged = True

    @builder
    def with_system_versioning(self) -> "CreateQueryBuilder":
        """
        Adds system versioning.

        :return:
            CreateQueryBuilder.
        """
        self._with_system_versioning = True

    @builder
    def columns(self, *columns: str | tuple[str, str] | Column) -> "CreateQueryBuilder":
        """
        Adds the columns.

        :param columns:
            Type:  Union[str, TypedTuple[str, str], Column]

            A list of columns.

        :raises AttributeError:
            If the table is an as_select table.

        :return:
            CreateQueryBuilder.
        """
        if self._as_select:
            raise AttributeError("'Query' object already has attribute as_select")

        for column in columns:
            if isinstance(column, str):
                column = Column(column)
            elif isinstance(column, tuple):
                column = Column(column_name=column[0], column_type=column[1])
            self._columns.append(column)

    @builder
    def period_for(
        self,
        name,
        start_column: str | Column,
        end_column: str | Column,
    ) -> "CreateQueryBuilder":
        """
        Adds a PERIOD FOR clause.

        :param name:
            The period name.

        :param start_column:
            The column that starts the period.

        :param end_column:
            The column that ends the period.

        :return:
            CreateQueryBuilder.
        """
        self._period_fors.append(PeriodFor(name, start_column, end_column))

    @builder
    def unique(self, *columns: str | Column) -> "CreateQueryBuilder":
        """
        Adds a UNIQUE constraint.

        :param columns:
            Type:  Union[str, Column]

            A list of columns.

        :return:
            CreateQueryBuilder.
        """
        self._uniques.append(self._prepare_columns_input(columns))

    @builder
    def primary_key(self, *columns: str | Column) -> "CreateQueryBuilder":
        """
        Adds a primary key constraint.

        :param columns:
            Type:  Union[str, Column]

            A list of columns.

        :raises AttributeError:
            If the primary key is already defined.

        :return:
            CreateQueryBuilder.
        """
        if self._primary_key:
            raise AttributeError("'Query' object already has attribute primary_key")
        self._primary_key = self._prepare_columns_input(columns)

    @builder
    def foreign_key(
        self,
        columns: list[str | Column],
        reference_table: str | Table,
        reference_columns: list[str | Column],
        on_delete: ReferenceOption = None,
        on_update: ReferenceOption = None,
    ) -> "CreateQueryBuilder":
        """
        Adds a foreign key constraint.

        :param columns:
            Type:  List[Union[str, Column]]

            A list of foreign key columns.

        :param reference_table:
            Type: Union[str, Table]

            The parent table name.

        :param reference_columns:
            Type: List[Union[str, Column]]

            Parent key columns.

        :param on_delete:
            Type: ReferenceOption

            Delete action.

        :param on_update:
            Type: ReferenceOption

            Update option.

        :raises AttributeError:
            If the foreign key is already defined.

        :return:
            CreateQueryBuilder.
        """
        if self._foreign_key:
            raise AttributeError("'Query' object already has attribute foreign_key")
        self._foreign_key = self._prepare_columns_input(columns)
        self._foreign_key_reference_table = reference_table
        self._foreign_key_reference = self._prepare_columns_input(reference_columns)
        self._foreign_key_on_delete = on_delete
        self._foreign_key_on_update = on_update

    @builder
    def as_select(self, query_builder: QueryBuilder) -> "CreateQueryBuilder":
        """
        Creates the table from a select statement.

        :param query_builder:
            The query.

        :raises AttributeError:
            If columns have been defined for the table.

        :return:
            CreateQueryBuilder.
        """
        if self._columns:
            raise AttributeError("'Query' object already has attribute columns")

        if not isinstance(query_builder, QueryBuilder):
            raise TypeError("Expected 'item' to be instance of QueryBuilder")

        self._as_select = query_builder

    @builder
    def if_not_exists(self) -> "CreateQueryBuilder":
        self._if_not_exists = True

    def get_sql(self, **kwargs: Any) -> str:
        """
        Gets the sql statement string.

        :return: The create table statement.
        :rtype: str
        """
        self._set_kwargs_defaults(kwargs)

        if not self._create_table:
            return ""

        if not self._columns and not self._as_select:
            return ""

        create_table = self._create_table_sql(**kwargs)

        if self._as_select:
            return create_table + self._as_select_sql(**kwargs)

        body = self._body_sql(**kwargs)
        table_options = self._table_options_sql(**kwargs)

        return f"{create_table} ({body}){table_options}"

    def _create_table_sql(self, **kwargs: Any) -> str:
        table_type = ''
        if self._temporary:
            table_type = 'TEMPORARY '
        elif self._unlogged:
            table_type = 'UNLOGGED '

        if_not_exists = ''
        if self._if_not_exists:
            if_not_exists = 'IF NOT EXISTS '

        return f"CREATE {table_type}TABLE {if_not_exists}{self._create_table.get_sql(**kwargs)}"

    def _table_options_sql(self, **kwargs) -> str:
        table_options = ""

        if self._with_system_versioning:
            table_options += ' WITH SYSTEM VERSIONING'

        return table_options

    def _column_clauses(self, **kwargs) -> list[str]:
        return [column.get_sql(**kwargs) for column in self._columns]

    def _period_for_clauses(self, **kwargs) -> list[str]:
        return [period_for.get_sql(**kwargs) for period_for in self._period_fors]

    def _unique_key_clauses(self, **kwargs) -> list[str]:
        return [
            "UNIQUE ({unique})".format(unique=",".join(column.get_name_sql(**kwargs) for column in unique))
            for unique in self._uniques
        ]

    def _primary_key_clause(self, **kwargs) -> str:
        return "PRIMARY KEY ({columns})".format(
            columns=",".join(column.get_name_sql(**kwargs) for column in self._primary_key),
        )

    def _foreign_key_clause(self, **kwargs) -> str:
        clause = "FOREIGN KEY ({columns}) REFERENCES {table_name} ({reference_columns})".format(
            columns=",".join(column.get_name_sql(**kwargs) for column in self._foreign_key),
            table_name=self._foreign_key_reference_table.get_sql(**kwargs),
            reference_columns=",".join(column.get_name_sql(**kwargs) for column in self._foreign_key_reference),
        )
        if self._foreign_key_on_delete:
            clause += " ON DELETE " + self._foreign_key_on_delete.value
        if self._foreign_key_on_update:
            clause += " ON UPDATE " + self._foreign_key_on_update.value

        return clause

    def _body_sql(self, **kwargs) -> str:
        clauses = self._column_clauses(**kwargs)
        clauses += self._period_for_clauses(**kwargs)
        clauses += self._unique_key_clauses(**kwargs)

        if self._primary_key:
            clauses.append(self._primary_key_clause(**kwargs))
        if self._foreign_key:
            clauses.append(self._foreign_key_clause(**kwargs))

        return ",".join(clauses)

    def _as_select_sql(self, **kwargs: Any) -> str:
        return f" AS ({self._as_select.get_sql(**kwargs)})"

    def _prepare_columns_input(self, columns: list[str | Column]) -> list[Column]:
        return [(column if isinstance(column, Column) else Column(column)) for column in columns]

    def __str__(self) -> str:
        return self.get_sql()

    def __repr__(self) -> str:
        return self.__str__()


class CreateIndexBuilder:
    def __init__(self) -> None:
        self._index = None
        self._columns = []
        self._table = None
        self._wheres = None
        self._is_unique = False
        self._if_not_exists = False

    @builder
    def create_index(self, index: str | Index) -> "CreateIndexBuilder":
        self._index = index

    @builder
    def columns(self, *columns: str | tuple[str, str] | Column) -> "CreateIndexBuilder":
        for column in columns:
            if isinstance(column, str):
                column = Column(column)
            elif isinstance(column, tuple):
                column = Column(column_name=column[0], column_type=column[1])
            self._columns.append(column)

    @builder
    def on(self, table: Table | str) -> "CreateIndexBuilder":
        self._table = table

    @builder
    def where(self, criterion: Term | EmptyCriterion) -> "CreateIndexBuilder":
        """
        Partial index where clause.
        """
        if self._wheres:
            self._wheres &= criterion
        else:
            self._wheres = criterion

    @builder
    def unique(self) -> "CreateIndexBuilder":
        self._is_unique = True

    @builder
    def if_not_exists(self) -> "CreateIndexBuilder":
        self._if_not_exists = True

    def get_sql(self) -> str:
        if not self._columns or len(self._columns) == 0:
            raise AttributeError("Cannot create index without columns")
        if not self._table:
            raise AttributeError("Cannot create index without table")
        columns_str = ", ".join([c.name for c in self._columns])
        unique_str = "UNIQUE" if self._is_unique else ""
        if_not_exists_str = "IF NOT EXISTS" if self._if_not_exists else ""
        base_sql = f"CREATE {unique_str} INDEX {if_not_exists_str} {self._index} ON {self._table}({columns_str})"
        if self._wheres:
            base_sql += f" WHERE {self._wheres}"
        return base_sql.replace("  ", " ")

    def __str__(self) -> str:
        return self.get_sql()

    def __repr__(self) -> str:
        return self.__str__()


class DropQueryBuilder:
    """
    Query builder used to build DROP queries.
    """

    QUOTE_CHAR = '"'
    SECONDARY_QUOTE_CHAR = "'"
    ALIAS_QUOTE_CHAR = None
    QUERY_CLS = Query

    def __init__(self, dialect: Dialects | None = None) -> None:
        self._drop_target_kind = None
        self._drop_target: Database | Table | str = ""
        self._if_exists = None
        self.dialect = dialect

    def _set_kwargs_defaults(self, kwargs: dict) -> None:
        kwargs.setdefault("quote_char", self.QUOTE_CHAR)
        kwargs.setdefault("secondary_quote_char", self.SECONDARY_QUOTE_CHAR)
        kwargs.setdefault("dialect", self.dialect)

    @builder
    def drop_database(self, database: Database | str) -> "DropQueryBuilder":
        target = database if isinstance(database, Database) else Database(database)
        self._set_target('DATABASE', target)

    @builder
    def drop_table(self, table: Table | str) -> "DropQueryBuilder":
        target = table if isinstance(table, Table) else Table(table)
        self._set_target('TABLE', target)

    @builder
    def drop_user(self, user: str) -> "DropQueryBuilder":
        self._set_target('USER', user)

    @builder
    def drop_view(self, view: str) -> "DropQueryBuilder":
        self._set_target('VIEW', view)

    @builder
    def drop_index(self, index: str) -> "DropQueryBuilder":
        self._set_target('INDEX', index)

    @builder
    def if_exists(self) -> "DropQueryBuilder":
        self._if_exists = True

    def _set_target(self, kind: str, target: Database | Table | str) -> None:
        if self._drop_target:
            raise AttributeError("'DropQuery' object already has attribute drop_target")
        self._drop_target_kind = kind
        self._drop_target = target

    def get_sql(self, **kwargs: Any) -> str:
        self._set_kwargs_defaults(kwargs)

        if_exists = 'IF EXISTS ' if self._if_exists else ''
        target_name: str = ""

        if isinstance(self._drop_target, (Database, Table)):
            target_name = self._drop_target.get_sql(**kwargs)
        else:
            target_name = format_quotes(self._drop_target, self.QUOTE_CHAR)

        return f"DROP {self._drop_target_kind} {if_exists}{target_name}"

    def __str__(self) -> str:
        return self.get_sql()

    def __repr__(self) -> str:
        return self.__str__()



import unittest

from pypika import Column, Columns, Query, Table, Tables
from pypika.enums import ReferenceOption
from pypika.terms import Index, ValueWrapper


class CreateTableTests(unittest.TestCase):
    new_table, existing_table = Tables("abc", "efg")
    foo, bar = Columns(("a", "INT"), ("b", "VARCHAR(100)"))

    def test_create_table_with_columns(self):
        with self.subTest("with nullable"):
            a = Column("a", "INT", True)
            b = Column("b", "VARCHAR(100)", False)
            q = Query.create_table(self.new_table).columns(a, b)

            self.assertEqual('CREATE TABLE "abc" ("a" INT NULL,"b" VARCHAR(100) NOT NULL)', str(q))

        with self.subTest("with defaults"):
            a = Column("a", "INT", default=ValueWrapper(42))
            b = Column("b", "VARCHAR(100)", default=ValueWrapper("foo"))
            q = Query.create_table(self.new_table).columns(a, b)

            self.assertEqual('CREATE TABLE "abc" ("a" INT DEFAULT 42,"b" VARCHAR(100) DEFAULT \'foo\')', str(q))

        with self.subTest("with unwrapped defaults"):
            a = Column("a", "INT", default=42)
            b = Column("b", "VARCHAR(100)", default="foo")
            q = Query.create_table(self.new_table).columns(a, b)

            self.assertEqual('CREATE TABLE "abc" ("a" INT DEFAULT 42,"b" VARCHAR(100) DEFAULT \'foo\')', str(q))

        with self.subTest("with period for"):
            a = Column("id", "INT")
            b = Column("valid_from", "DATETIME")
            c = Column("valid_to", "DATETIME")
            q = Query.create_table(self.new_table).columns(a, b, c).period_for('valid_period', b, c)

            self.assertEqual(
                'CREATE TABLE "abc" ('
                '"id" INT,'
                '"valid_from" DATETIME,'
                '"valid_to" DATETIME,'
                'PERIOD FOR "valid_period" ("valid_from","valid_to"))',
                str(q),
            )

        with self.subTest("without temporary keyword"):
            q = Query.create_table(self.new_table).columns(self.foo, self.bar)

            self.assertEqual('CREATE TABLE "abc" ("a" INT,"b" VARCHAR(100))', str(q))

        with self.subTest("with temporary keyword"):
            q = Query.create_table(self.new_table).temporary().columns(self.foo, self.bar)

            self.assertEqual('CREATE TEMPORARY TABLE "abc" ("a" INT,"b" VARCHAR(100))', str(q))

        with self.subTest("with primary key"):
            q = Query.create_table(self.new_table).columns(self.foo, self.bar).primary_key(self.foo, self.bar)

            self.assertEqual('CREATE TABLE "abc" ("a" INT,"b" VARCHAR(100),PRIMARY KEY ("a","b"))', str(q))

        with self.subTest("with simple foreign key"):
            cref, dref = Columns(("c", "INT"), ("d", "VARCHAR(100)"))
            q = (
                Query.create_table(self.new_table)
                .columns(self.foo, self.bar)
                .foreign_key([self.foo, self.bar], self.existing_table, [cref, dref])
            )

            self.assertEqual(
                'CREATE TABLE "abc" ("a" INT,"b" VARCHAR(100),FOREIGN KEY ("a","b") REFERENCES "efg" ("c","d"))',
                str(q),
            )

        with self.subTest("with foreign key reference options"):
            cref, dref = Columns(("c", "INT"), ("d", "VARCHAR(100)"))
            q = (
                Query.create_table(self.new_table)
                .columns(self.foo, self.bar)
                .foreign_key(
                    [self.foo, self.bar],
                    self.existing_table,
                    [cref, dref],
                    on_delete=ReferenceOption.cascade,
                    on_update=ReferenceOption.restrict,
                )
            )

            self.assertEqual(
                'CREATE TABLE "abc" ('
                '"a" INT,'
                '"b" VARCHAR(100),'
                'FOREIGN KEY ("a","b") REFERENCES "efg" ("c","d") ON DELETE CASCADE ON UPDATE RESTRICT)',
                str(q),
            )

        with self.subTest("with unique keys"):
            q = (
                Query.create_table(self.new_table)
                .columns(self.foo, self.bar)
                .unique(self.foo, self.bar)
                .unique(self.foo)
            )

            self.assertEqual('CREATE TABLE "abc" ("a" INT,"b" VARCHAR(100),UNIQUE ("a","b"),UNIQUE ("a"))', str(q))

        with self.subTest("with system versioning"):
            q = Query.create_table(self.new_table).columns(self.foo, self.bar).with_system_versioning()

            self.assertEqual('CREATE TABLE "abc" ("a" INT,"b" VARCHAR(100)) WITH SYSTEM VERSIONING', str(q))

        with self.subTest("with unlogged keyword"):
            q = Query.create_table(self.new_table).unlogged().columns(self.foo, self.bar)

            self.assertEqual('CREATE UNLOGGED TABLE "abc" ("a" INT,"b" VARCHAR(100))', str(q))

        with self.subTest("with if not exists keyword"):
            q = Query.create_table(self.new_table).if_not_exists().columns(self.foo, self.bar)

            self.assertEqual('CREATE TABLE IF NOT EXISTS "abc" ("a" INT,"b" VARCHAR(100))', str(q))

    def test_create_table_with_select(self):
        select = Query.from_(self.existing_table).select(self.existing_table.foo, self.existing_table.bar)

        with self.subTest("without temporary keyword"):
            q = Query.create_table(self.new_table).as_select(select)

            self.assertEqual('CREATE TABLE "abc" AS (SELECT "foo","bar" FROM "efg")', str(q))

        with self.subTest("with temporary keyword"):
            q = Query.create_table(self.new_table).temporary().as_select(select)

            self.assertEqual(
                'CREATE TEMPORARY TABLE "abc" AS (SELECT "foo","bar" FROM "efg")',
                str(q),
            )

    def test_create_table_without_columns_or_select_empty(self):
        q = Query.create_table(self.new_table)

        self.assertEqual("", str(q))

    def test_create_table_with_select_and_columns_fails(self):
        select = Query.from_(self.existing_table).select(self.existing_table.foo, self.existing_table.bar)

        with self.subTest("for columns before as_select"), self.assertRaises(AttributeError):
            Query.create_table(self.new_table).columns(self.foo, self.bar).as_select(select)

        with self.subTest("for as_select before columns"), self.assertRaises(AttributeError):
            Query.create_table(self.new_table).as_select(select).columns(self.foo, self.bar)

        with self.subTest("repeated foreign key"), self.assertRaises(AttributeError):
            Query.create_table(self.new_table).foreign_key([self.foo], self.existing_table, [self.bar]).foreign_key(
                [self.foo],
                self.existing_table,
                [self.bar],
            )

    def test_create_table_as_select_not_query_raises_error(self):
        with self.assertRaises(TypeError):
            Query.create_table(self.new_table).as_select("abc")


class CreateIndexTests(unittest.TestCase):
    new_index = Index("abc")
    new_table = Table("my_table")
    columns = (new_table.a, new_table.b)
    where = new_table.a > 1

    def test_create_index_simple(self) -> None:
        q = Query.create_index(self.new_index).on(self.new_table).columns(*self.columns)
        self.assertEqual(
            f'CREATE INDEX "{self.new_index.name}" ON "{self.new_table.get_table_name()}"({", ".join([c.name for c in self.columns])})',
            q.get_sql(),
        )

    def test_create_index_where(self) -> None:
        q = Query.create_index(self.new_index).on(self.new_table).columns(*self.columns).where(self.where)
        self.assertEqual(
            f'CREATE INDEX "{self.new_index.name}" ON "{self.new_table.get_table_name()}"({", ".join([c.name for c in self.columns])}) WHERE {self.where.get_sql()}',
            q.get_sql(),
        )

    def test_create_index_unique(self) -> None:
        q = Query.create_index(self.new_index).on(self.new_table).columns(*self.columns).unique()
        self.assertEqual(
            f'CREATE UNIQUE INDEX "{self.new_index.name}" ON "{self.new_table.get_table_name()}"({", ".join([c.name for c in self.columns])})',
            q.get_sql(),
        )

    def test_create_index_if_not_exists(self) -> None:
        q = Query.create_index(self.new_index).on(self.new_table).columns(*self.columns).if_not_exists()
        self.assertEqual(
            f'CREATE INDEX IF NOT EXISTS "{self.new_index.name}" ON "{self.new_table.get_table_name()}"({", ".join([c.name for c in self.columns])})',
            q.get_sql(),
        )

    def test_create_index_without_columns_raises_error(self) -> None:
        with self.assertRaises(AttributeError):
            Query.create_index(self.new_index).on(self.new_table).get_sql()

    def test_create_index_without_table_raises_error(self) -> None:
        with self.assertRaises(AttributeError):
            Query.create_index(self.new_index).columns(*self.columns).get_sql()



import abc

from pypika.terms import Function
from pypika.utils import format_alias_sql


class _AbstractSearchString(Function, metaclass=abc.ABCMeta):
    def __init__(self, name, pattern: str, alias: str | None = None):
        super(_AbstractSearchString, self).__init__(self.clickhouse_function(), name, alias=alias)

        self._pattern = pattern

    @classmethod
    @abc.abstractmethod
    def clickhouse_function(cls) -> str:
        pass

    def get_sql(self, with_alias=False, with_namespace=False, quote_char=None, dialect=None, **kwargs):
        args = []
        for p in self.args:
            if hasattr(p, "get_sql"):
                args.append(f'toString("{p.get_sql(with_alias=False, **kwargs)}")')
            else:
                args.append(str(p))

        sql = "{name}({args},'{pattern}')".format(
            name=self.name,
            args=",".join(args),
            pattern=self._pattern,
        )
        return format_alias_sql(sql, self.alias, **kwargs)


class Match(_AbstractSearchString):
    @classmethod
    def clickhouse_function(cls) -> str:
        return "match"


class Like(_AbstractSearchString):
    @classmethod
    def clickhouse_function(cls) -> str:
        return "like"


class NotLike(_AbstractSearchString):
    @classmethod
    def clickhouse_function(cls) -> str:
        return "notLike"


class _AbstractMultiSearchString(Function, metaclass=abc.ABCMeta):
    def __init__(self, name, patterns: list, alias: str | None = None):
        super(_AbstractMultiSearchString, self).__init__(self.clickhouse_function(), name, alias=alias)

        self._patterns = patterns

    @classmethod
    @abc.abstractmethod
    def clickhouse_function(cls) -> str:
        pass

    def get_sql(self, with_alias=False, with_namespace=False, quote_char=None, dialect=None, **kwargs):
        args = []
        for p in self.args:
            if hasattr(p, "get_sql"):
                args.append(f'toString("{p.get_sql(with_alias=False, **kwargs)}")')
            else:
                args.append(str(p))

        sql = "{name}({args},[{patterns}])".format(
            name=self.name,
            args=",".join(args),
            patterns=",".join(["'%s'" % i for i in self._patterns]),
        )
        return format_alias_sql(sql, self.alias, **kwargs)


class MultiSearchAny(_AbstractMultiSearchString):
    @classmethod
    def clickhouse_function(cls) -> str:
        return "multiSearchAny"


class MultiMatchAny(_AbstractMultiSearchString):
    @classmethod
    def clickhouse_function(cls) -> str:
        return "multiMatchAny"
